"""
Created on Tue Jul 24 11:47:57 2012

@author: eendebakpt
"""

# %% Load necessary packages """
import sys

import matplotlib.pyplot as plt
import numpy as np

import oapackage


def addData(data, extra_data, index=0, arrays=None, extra_arrays=None):
    """Vertically stack data together"""
    nn = extra_data.shape[1]
    ww = np.vstack(
        (
            extra_data,
            index
            * np.ones(
                nn,
            ),
        )
    )
    data = np.hstack((data, ww))
    if arrays is None:
        return data
    else:
        allarrays = oapackage.joinArrayLists([arrays, extra_arrays])
        return data, allarrays


# %% Load a strength 3 orthogonal array that is D-optimal


strength3_pareto = oapackage.exampleArray(44)
arrayclass = oapackage.arraylink2arraydata(strength3_pareto)

# %% Optimize with alpha variations


def optimize_arrays(sols, alpha=None, verbose=1, niter=400):
    """Optimize set of designs"""
    vv2 = []
    ll = oapackage.arraylist_t()
    for ii, A in enumerate(sols):
        Dinit = A.Defficiency()
        Df, Ax = oapackage.optimDeffPython(A, niter=niter, verbose=0, alpha=alpha)
        D, Ds, D1 = Ax.Defficiencies()
        ll.push_back(Ax)
        oapackage.tprint("optimize array %d: %f -> D %f Ds %f D1 %f" % (ii, Dinit, D, Ds, D1))
        vv2.append(np.array([D, Ds, D1]))
    return np.array(vv2).T, ll


def getOptimizationResults(arrayclass, alpha, nrestarts=80):
    scores, dds, designs, _ = oapackage.Doptimize(arrayclass, nrestarts=nrestarts, optimfunc=alpha, selectpareto=False)
    if 1:
        dds_extra, designs_extra = optimize_arrays([strength3_pareto, strength3_pareto, strength3_pareto], alpha=alpha)
        dds = np.vstack((dds, dds_extra))
        designs = designs + list(designs_extra)
    dds = np.array([al.Defficiencies() for al in designs])
    return dds, designs


optim_functions = [[1, 0, 0], [1, 1, 0], [1, 3, 0]]

results = []
oapackage.seedfastrand(600)
oapackage.set_srand(50)
for jj, optim_func in enumerate(optim_functions):
    print("optimize series %d: optim_func %s" % (jj, optim_func))
    sys.stdout.flush()
    dds, arrays = getOptimizationResults(arrayclass, alpha=optim_func)
    results.append((dds, arrays, optim_func))


# %% Combine the data

data = np.zeros((4, 0))
allarrays = oapackage.arraylist_t()
lbls = []

for jj, result in enumerate(results):
    dds, arrays, optim_func = result
    data, allarrays = addData(data, dds.T, jj, allarrays, arrays)
    lbl = "Optimization of $D$"
    if optim_func[1] > 0:
        if optim_func[1] == 1:
            lbl += " + $D_s$"
        else:
            lbl += f" + {optim_func[1]:g}$D_s$"
    if optim_func[2] > 0:
        lbl += f" + {optim_func[2]:g}$D_{{s0}}$"
    lbls += [lbl]

data = data[[0, 1, -1], :]
x = oapackage.generateDscatter(data.T, lbls=lbls, fig=30, ndata=data.shape[0] - 1, scatterarea=80)
oapackage.oahelper.setWindowRectangle(1900, 10, w=900, h=600)

ddst3 = np.array([al.Defficiencies() for al in [strength3_pareto]])


if arrayclass.N == 40:
    plt.gca().scatter(
        ddst3[:, 1], ddst3[:, 0], marker="s", s=90, c=(0, 0, 0), linewidths=0, alpha=1, label="Strength 3"
    )
    xl = plt.xlim()
    plt.xlim([xl[0], 1.05])
    yl = plt.ylim()
    plt.ylim([0.78, yl[1]])
    plt.draw()
    plt.show()
    plt.legend(numpoints=1, scatterpoints=1, loc=3)

# researchOA.niceplot(plt.gca())
# plt.savefig(os.path.join(xdir, 'motivating-%s-scatterplot-ndata%d.png' %
#                         (arrayclass.idstr().replace('.', '-d-'), data.shape[0] - 1)), dpi=150)



"""Module to work with designs and graphs

@author: Pieter Eendebak <pieter.eendebak@gmail.com>
"""

# %% Load packages
from typing import Any

import numpy as np

import oapackage

# %%


def oa2graph(al, adata, verbose=1):
    """
    Convert orthogonal array to graph representation

    The conversion method is as in Ryan and Bulutoglu. The resulting graph is bi-partite.
    The graph representation can be used for isomorphism testing.

    Example
    -------
    >     im, colors, r = oa2graph( A, oadata );

    """
    A = al.getarray(verbose=0)
    if verbose:
        print(f"oa2graph: array of shape {A.shape}")
    nrows = adata.N
    ncols = A.shape[1]
    nColumnLevelVertices = sum(adata.factor_levels())
    nVertices = adata.N + ncols + nColumnLevelVertices
    nColVertices = ncols
    colOffset = adata.N

    s = np.array(adata.factor_levels())
    sc = np.cumsum(s)
    sc0 = np.hstack(([0], sc))
    qq = np.ones(nColumnLevelVertices)  # colors for column level vertices
    for ii in range(sc.size):
        qq[sc0[ii] : sc0[ii + 1]] = 2 + ii
    qq = 2 * np.ones(nColumnLevelVertices)  # colors for column level vertices

    vertexOffsets = adata.N + ncols + np.hstack((0, s[0:-1])).cumsum()
    colors = np.hstack((np.zeros(adata.N), np.ones(ncols), qq))

    im = np.zeros((nVertices, nVertices))  # incidence matrix

    for row in range(0, nrows):
        idx = A[row, :] + vertexOffsets
        im[row, idx] = 1
        im[idx, row] = 1

    if nColVertices > 0:
        colidx = 2
        for col in range(0, ncols):
            sel = vertexOffsets[col] + range(0, s[col])
            im[colOffset + col, sel] = colidx
            im[sel, colOffset + col] = colidx

    # The non-row vertices do not have any connections to other non-row
    # vertices.
    return im, colors, dict({"adata": adata, "im": im, "colors": colors, "nVertices": nVertices})


def selectIsomorphismClasses(sols, verbose: int = 1) -> tuple[list[int], list[Any]]:
    """Select isomorphism classes from a list of designs

    Args:
        sols (list of arrays): list of arrays from which to determine the unique ones
        verbose (int): verbosity level
    Return:
        indices (list of integers): indices of the isomorphism classes
        mm (list of arrays): the arrays in normal form


    Example:
        >>> import oapackage.graphtools; import numpy as np
        >>> sols=[oapackage.exampleArray(idx) for idx in [26,26,27,26]]
        >>> idx, mm = oapackage.graphtools.selectIsomorphismClasses(sols, verbose=0)
        >>> _, unique_indices = np.unique(idx, return_index=True)
        >>> print('found %d isomorphism classes from %d arrays' % (len(unique_indices), len(sols)) )
        found 2 isomorphism classes from 4 arrays
        >>> representatives = [sols[idx] for idx in unique_indices]

    """

    mm = []
    for ii, al in enumerate(sols):
        if verbose:
            oapackage.tprint("selectIsomorphismClasses: process aray %d/%d" % (ii, len(sols)))
        al = oapackage.makearraylink(al)

        tt = oapackage.reduceOAnauty(al, verbose >= 2)

        alx = tt.apply(al)
        mm.append(np.array(alx))

    # perform uniqueness check
    nn = len(mm)
    qq = np.array([None] * nn, dtype=object)
    for ii in range(nn):
        qq[ii] = mm[ii].flatten()

    # Trick to make unique work...
    _, indices = np.unique(np.vstack(qq), axis=0, return_inverse=True)  # ty: ignore

    if verbose >= 1:
        print(f"selectIsomorphismClasses: reduce {len(sols)} to {np.unique(indices).size}")

    return indices, mm



import re
import sys

""" https://github.com/joshburnett/scanf

This file is extracted from https://github.com/joshburnett/scanf. The code is protected by an MIT license.

"""
"""
Small scanf implementation.

Python has powerful regular expressions but sometimes they are totally overkill
when you just want to parse a simple-formatted string.
C programmers use the scanf-function for these tasks (see link below).

This implementation of scanf translates the simple scanf-format into
regular expressions. Unlike C you can be sure that there are no buffer overflows
possible.

For more information see
  * http://www.python.org/doc/current/lib/node49.html
  * http://en.wikipedia.org/wiki/Scanf

Original code from:
    http://code.activestate.com/recipes/502213-simple-scanf-implementation/

Modified original to make the %f more robust, as well as added %* modifier to
skip fields.
"""
try:
    from functools import lru_cache
except ImportError:
    from backports.functools_lru_cache import lru_cache  # type: ignore

__version__ = "1.5.2"

__all__ = ["scanf", "extractdata", "scanf_translate", "scanf_compile"]


DEBUG = False


# As you can probably see it is relatively easy to add more format types.
# Make sure you add a second entry for each new item that adds the extra
#   few characters needed to handle the field ommision.
scanf_translate = [
    (re.compile(_token), _pattern, _cast)
    for _token, _pattern, _cast in [
        (r"%c", r"(.)", lambda x: x),
        (r"%\*c", r"(?:.)", None),
        (r"%(\d)c", r"(.{%s})", lambda x: x),
        (r"%\*(\d)c", r"(?:.{%s})", None),
        (r"%(\d)[di]", r"([+-]?\d{%s})", int),
        (r"%\*(\d)[di]", r"(?:[+-]?\d{%s})", None),
        ("%[di]", r"([+-]?\d+)", int),
        (r"%\*[di]", r"(?:[+-]?\d+)", None),
        ("%u", r"(\d+)", int),
        (r"%\*u", r"(?:\d+)", None),
        (r"%[fgeE]", r"([-+]?(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][-+]?\d+)?)", float),
        (r"%\*[fgeE]", r"(?:[-+]?(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][-+]?\d+)?)", None),
        (r"%s", r"(\S+)", lambda x: x),
        (r"%\*s", r"(?:\S+)", None),
        (r"%([xX])", r"(0%s[\dA-Za-f]+)", lambda x: int(x, 16)),
        (r"%\*([xX])", r"(?:0%s[\dA-Za-f]+)", None),
        (r"%o", r"(0[0-7]*)", lambda x: int(x, 8)),
        (r"%\*o", r"(?:0[0-7]*)", None),
    ]
]


# Cache formats
SCANF_CACHE_SIZE = 1000


@lru_cache(maxsize=SCANF_CACHE_SIZE)
def scanf_compile(format, collapseWhitespace=True):
    r"""
    Translate the format into a regular expression

    For example:
    >>> format_re, casts = scanf_compile('%s - %d errors, %d warnings')
    >>> print format_re.pattern
    (\S+) \- ([+-]?\d+) errors, ([+-]?\d+) warnings

    Translated formats are cached for faster reuse
    """

    format_pat = ""
    cast_list = []
    i = 0
    length = len(format)
    while i < length:
        found = None
        for token, pattern, cast in scanf_translate:
            found = token.match(format, i)
            if found:
                if cast:  # cast != None
                    cast_list.append(cast)
                groups = found.groupdict() or found.groups()
                if groups:
                    pattern = pattern % groups
                format_pat += pattern
                i = found.end()
                break
        if not found:
            char = format[i]
            # escape special characters
            if char in "|^$()[]-.+*?{}<>\\":
                format_pat += "\\"
            format_pat += char
            i += 1
    if DEBUG:
        print(f"DEBUG: {format!r} -> {format_pat}")
    if collapseWhitespace:
        format_pat = re.sub(r"\s+", r"\\s+", format_pat)

    format_re = re.compile(format_pat)
    return format_re, cast_list


def scanf(format, s=None, collapseWhitespace=True):
    """
    scanf supports the following formats:
      %c        One character
      %5c       5 characters
      %d, %i    int value
      %7d, %7i  int value with length 7
      %f        float value
      %o        octal value
      %X, %x    hex value
      %s        string terminated by whitespace

    Examples:
    >>> scanf("%s - %d errors, %d warnings", "/usr/sbin/sendmail - 0 errors, 4 warnings")
    ('/usr/sbin/sendmail', 0, 4)
    >>> scanf("%o %x %d", "0123 0x123 123")
    (83, 291, 123)


    scanf.scanf returns a tuple of found values
    or None if the format does not match.
    """

    if s is None:
        s = sys.stdin

    if hasattr(s, "readline"):
        s = s.readline()

    format_re, casts = scanf_compile(format, collapseWhitespace)

    found = format_re.search(s)
    if found:
        groups = found.groups()
        return tuple([casts[i](groups[i]) for i in range(len(groups))])


def extractdata(pattern, text=None, filepath=None):
    """
    Read through an entire file or body of text one line at a time. Parse each line that matches the supplied
    pattern string and ignore the rest.

    If *text* is supplied, it will be parsed according to the *pattern* string.
    If *text* is not supplied, the file at *filepath* will be opened and parsed.
    """
    y = []
    if text is None:
        textsource = open(filepath)
    else:
        textsource = text.splitlines()

    for line in textsource:
        match = scanf(pattern, line)
        if match:
            if len(y) == 0:
                y = [[s] for s in match]
            else:
                for i, ydata in enumerate(y):
                    ydata.append(match[i])

    if text is None:
        textsource.close()

    return y


if __name__ == "__main__":
    import doctest

    doctest.testmod(verbose=True, report=True)



#!/usr/bin/env python3
#
# oapackage documentation build configuration file, created by
# sphinx-quickstart on Tue Jun 26 20:08:57 2018.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# %% The oapackage cannot be build on rtd. We mock the module
import os
import platform
import subprocess

rtd = os.environ.get("READTHEDOCS", False)
mock_oalib = False

if rtd:
    import sys
    from unittest.mock import MagicMock

    class Mock(MagicMock):
        @classmethod
        def __getattr__(cls, name):
            return MagicMock()

    if mock_oalib:
        MOCK_MODULES = ["_oalib"]
        sys.modules.update((mod_name, Mock()) for mod_name in MOCK_MODULES)

    print("##############################")
    print(f"current directory: {os.getcwd()}")
    print(sorted(os.listdir()))
    print("##############################")
    sys.path.append(".")
    print("##############################")

if rtd:
    subprocess.call("cd ../; doxygen Doxyfile; python doxy2swig.py -a -c docs/xml/index.xml oadoxy.i", shell=True)
else:
    print("executing doxygen")
    if platform.system() == "Windows":
        r = subprocess.call("cd ../ && doxygen Doxyfile", shell=True)
    else:
        r = subprocess.call("cd ../; doxygen Doxyfile", shell=True)
    print("executing doxygen done")

# %%
# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    #              'sphinx_rtd_theme',
    "sphinx.ext.mathjax",
]
extensions += ["sphinxcontrib.bibtex"]
extensions += ["nbsphinx"]
# extensions += ['sphinxcontrib.napoleon']
extensions += ["sphinx.ext.napoleon"]
extensions += ["sphinx.ext.intersphinx"]
extensions += ["sphinx.ext.autosectionlabel"]
extensions += ["IPython.sphinxext.ipython_console_highlighting"]
extensions += ["sphinx.ext.autosummary"]
extensions += ["sphinx.ext.doctest"]

bibtex_bibfiles = ["references.bib"]

extensions += ["breathe"]
breathe_projects = {"oapackage": "./xml"}

breathe_default_project = "oapackage"


# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "oapackage"
copyright = "2018, Pieter Eendebak, Alan Vazquez"
author = "Pieter Eendebak, Alan Vazquez"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#

# import oapackage
# version = '{}'.format(oapackage.__version__)
# The short X.Y version.
version = ""
# The full version, including alpha/beta/rc tags.
release = version

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = [
    "setup.py",
    "_build",
    "Thumbs.db",
    ".DS_Store",
    "../oapackage/tests.py",
    "../oapackage/markup.py",
    "examples/.ipynb_checkpoints/*",
]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "alabaster"
html_theme = "sphinx_rtd_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "oapackagedoc"


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, "oapackage.tex", "oapackage Documentation", "Pieter Eendebak, Alan Vazquez", "manual"),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "oapackage", "oapackage Documentation", [author], 1)]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "oapackage",
        "oapackage Documentation",
        author,
        "oapackage",
        "One line description of project.",
        "Miscellaneous",
    ),
]

# %%


def run_apidoc(_):
    import os

    print(f"run_apidoc: current dir is {os.getcwd()}")

    ignore_paths = [
        os.path.join("oapackage", "markup.py"),
        "get_artifacts.py",
        "untitled*.py",
        "doxy2swig.py",
        "setup.py",
    ]
    ignore_paths = [os.path.join("..", file) for file in ignore_paths]

    argv = [
        "-f",
        #        "-T",
        #        "-e",
        "-M",
        "-o",
        ".",
        os.path.join(".."),
        # os.path.join("..", 'oapackage')
    ] + ignore_paths

    sphinxcmd = "sphinx-apidoc " + " ".join(argv)
    if rtd:
        print(sphinxcmd)

    if 1:
        try:
            # Sphinx 1.7+
            from sphinx.ext import apidoc

            apidoc.main(argv)
        except ImportError:
            # Sphinx 1.6 (and earlier)
            from sphinx import apidoc

            argv.insert(0, apidoc.__file__)
            apidoc.main(argv)
    else:
        subprocess.call("dir", shell=True)
        subprocess.call(sphinxcmd, shell=True)


def setup(app):
    print("conf.py: setup")
    app.connect("builder-inited", run_apidoc)



import unittest

import oapackage


class TestGWLP(unittest.TestCase):
    def test_gwlp(self):
        array = oapackage.exampleArray(1, 0)
        array = array.selectFirstColumns(3)
        gwlp = oapackage.GWLP(array)
        gwlp_mixed = oapackage.GWLPmixed(array, 0)

        self.assertEqual(gwlp, (1.0, 0.0, 0.0, 0.25))
        self.assertEqual(gwlp, gwlp_mixed)

    def test_gwlp_mixed(self):
        array = oapackage.exampleArray(56, 0)
        gwlp = oapackage.GWLPmixed(array, 0)
        self.assertEqual(
            gwlp,
            (
                1.0,
                0.0,
                0.0,
                48.0,
                185.0,
                524.0,
                1381.0,
                2792.0,
                4327.0,
                5556.0,
                5961.0,
                5152.0,
                3539.0,
                1972.0,
                911.0,
                328.0,
                76.0,
                12.0,
                3.0,
                0.0,
            ),
        )

    def test_gwlp_mixed2(self):
        al = oapackage.exampleArray(4, 0)
        self.assertEqual(al.GWLP(), (1.0, 0.0, 0.0, 3.5, 2.5, 0.5, 0.5, 0.0))

    def test_regression_gwlp_non_mixed(self):
        al = oapackage.exampleArray(57, 0)
        gwlp = al.GWLP()
        self.assertEqual(gwlp, (1.0, 0.0, 0.0, 13.0, 13.5, 9.0, 4.0))
        self.assertEqual(oapackage.GWLPmixed(al), (1.0, 0.0, 0.0, 13.0, 13.5, 9.0, 4.0))


if __name__ == "__main__":
    """ Test code """
    unittest.main()



"""Orthogonal Array package test functions"""

import io
import os
import tempfile
import time
import unittest
import unittest.mock as mock
from contextlib import redirect_stdout
from unittest.mock import patch

import numpy as np

import oapackage


def is_sorted(lst):
    return all(a <= b for a, b in zip(lst, lst[1:]))


class TestCompareMethods(unittest.TestCase):
    def test_compareLMC(self):
        al1 = oapackage.exampleArray(24, 0)
        al2 = oapackage.exampleArray(25, 0)
        al3 = oapackage.exampleArray(26, 0)

        self.assertEqual(oapackage.compareLMC(al1, al2), -1)
        self.assertEqual(oapackage.compareLMC(al2, al1), 1)
        self.assertEqual(oapackage.compareLMC(al1, al3), 1)

        al1 = oapackage.exampleArray(2, 0).selectFirstColumns(5)
        al2 = oapackage.exampleArray(2, 0).selectLastColumns(5)
        self.assertEqual(oapackage.compareLMC(al1, al2), -1)

    def test_compareJ54(self):
        al1 = oapackage.exampleArray(24, 0)
        al2 = oapackage.exampleArray(25, 0)
        al3 = oapackage.exampleArray(26, 0)

        self.assertEqual(oapackage.compareJ54(al1, al2), -1)
        self.assertEqual(oapackage.compareJ54(al2, al1), 1)

        self.assertEqual(oapackage.compareJ54(al1, al3), 1)
        self.assertEqual(oapackage.compareJ54(al2, al3), 1)


class TestMinimalFormCheck(unittest.TestCase):
    def test_LMCcheck(self):
        array = oapackage.exampleArray(1)

        lmc_type = oapackage.LMCcheck(array)
        self.assertEqual(lmc_type, oapackage.LMC_MORE)

        array2 = array.selectColumns([0, 1, 2, 4, 3])
        lmc_type = oapackage.LMCcheck(array2)
        self.assertEqual(lmc_type, oapackage.LMC_LESS)

        array2 = array.selectColumns([2, 1, 0, 3, 4])
        lmc_type = oapackage.LMCcheck(array2)
        self.assertEqual(lmc_type, oapackage.LMC_LESS)

        arrayclass = oapackage.arraylink2arraydata(array)
        reduction = oapackage.LMCreduction_t(arrayclass)
        oaextend = oapackage.OAextend(arrayclass)
        lmc_type = oapackage.LMCcheck(array2.array, arrayclass, oaextend, reduction)
        self.assertEqual(lmc_type, oapackage.LMC_LESS)

    def test_LMCcheckOriginal(self):
        array = oapackage.exampleArray(1)
        lmc_type = oapackage.LMCcheckOriginal(array)
        self.assertEqual(lmc_type, oapackage.LMC_MORE)

        array2 = array.selectColumns([0, 1, 2, 4, 3])
        lmc_type = oapackage.LMCcheck(array2)
        self.assertEqual(lmc_type, oapackage.LMC_LESS)

        array = oapackage.exampleArray(5, 0)
        with self.assertRaises(RuntimeError):
            lmc_type = oapackage.LMCcheckOriginal(array)


class TestReductions(unittest.TestCase):
    def test_LMC(self):
        al = oapackage.array_link(2, 2, 0)
        al[1, 1] = -1
        self.assertRaises(RuntimeError, oapackage.reduceLMCform, al)

        al = oapackage.exampleArray(8, 0)
        alr = oapackage.reduceLMCform(al)
        self.assertTrue(alr == al)

        array0 = oapackage.exampleArray(1, 0).selectFirstColumns(3)
        array = array0.randomperm()
        reduced_array = oapackage.reduceLMCform(array)
        self.assertEqual(array0, reduced_array)

    def test_DOP(self):
        al = oapackage.exampleArray(1, 0)
        transformation = oapackage.reductionDOP(al)
        self.assertTrue(transformation.isIdentity())

        dof_values = oapackage.projectionDOFvalues(al)
        pvalues = [pvalue.raw_values() for pvalue in dof_values]
        self.assertTrue(is_sorted(pvalues))

        al = oapackage.exampleArray(7, 0)
        transformation = oapackage.reductionDOP(al)
        self.assertTrue(transformation.isIdentity())

        al = oapackage.exampleArray(8, 0)
        transformation = oapackage.reductionDOP(al)

        with mock.patch("sys.stdout", new_callable=io.StringIO) as mock_stdout:
            transformation.show()
            stdout = mock_stdout.getvalue()
            self.assertTrue(stdout.startswith("array transformation: N 40"))
            self.assertIn("column permutation: {2,0,3,1,4,5,6}", stdout)
        alr = oapackage.reduceDOPform(al)
        self.assertTrue(transformation.apply(al) == alr)


class TestModelmatrix(unittest.TestCase):
    def test_modelmatrix_main_effects(self):
        # test model matrix main effects are helmert contrasts
        array = oapackage.array_link(np.array([[0, 1, 2, 3]]).T)

        M = oapackage.array2modelmatrix(array, "m")
        hc = oapackage.oahelper.helmert_contrasts(4)
        np.testing.assert_array_almost_equal(hc, M[:, 1:])

    def test_modelmatrix(self):
        al = oapackage.exampleArray(1, 0)

        sizes = oapackage.array2modelmatrix_sizes(al)
        k = al.n_columns
        self.assertEqual(sizes, (1, 1 + k, int(1 + k + k * (k - 1) / 2), int(1 + k + k * (k + 1) / 2)))

        model_matrix = oapackage.array2modelmatrix(al, "main", verbose=0)
        np.testing.assert_array_equal(2 * np.array(al) - 1, model_matrix[:, 1:])

        conf_design = oapackage.exampleArray(41, 0)
        sizes = oapackage.array2modelmatrix_sizes(conf_design)
        k = conf_design.n_columns
        self.assertEqual(sizes, (1, 1 + k, int(1 + k + k * (k - 1) / 2), int(1 + k + k * (k + 1) / 2)))
        model_matrix = oapackage.array2modelmatrix(conf_design, "i", 0)
        self.assertTrue(model_matrix.shape[1] == sizes[2])
        model_matrix = oapackage.array2modelmatrix(conf_design, "q", 0)

        last_column = np.array(conf_design)[:, -1]

        np.testing.assert_array_equal(model_matrix[:, -1], last_column * last_column)

        self.assertTrue(model_matrix.shape[1] == sizes[3])

    def test_array2eigenModelMatrixMixed(self):
        array = oapackage.exampleArray(0, 0)
        r = oapackage.array2eigenModelMatrixMixed(array.selectFirstColumns(1), verbose=0)
        main_effects = r[0]
        np.testing.assert_array_equal(main_effects, np.array([[-1.0, -1.0, -1.0, -1.0, 1.0, 1.0, 1.0, 1.0]]).T)

        array = oapackage.exampleArray(10, 0).selectFirstColumns(2)
        r = oapackage.array2eigenModelMatrixMixed(array, verbose=0)
        main_effects = r[0]
        interaction_model = r[1]
        np.testing.assert_array_almost_equal(
            main_effects,
            np.array(
                [
                    [-1.22474487, -0.70710678, -1.22474487, -0.70710678],
                    [-1.22474487, -0.70710678, 1.22474487, -0.70710678],
                    [-1.22474487, -0.70710678, 0.0, 1.41421356],
                    [1.22474487, -0.70710678, 1.22474487, -0.70710678],
                    [1.22474487, -0.70710678, 1.22474487, -0.70710678],
                    [1.22474487, -0.70710678, 0.0, 1.41421356],
                    [0.0, 1.41421356, -1.22474487, -0.70710678],
                    [0.0, 1.41421356, -1.22474487, -0.70710678],
                    [0.0, 1.41421356, 0.0, 1.41421356],
                ]
            ),
        )
        np.testing.assert_array_almost_equal(
            interaction_model,
            np.array(
                [
                    [1.5, 0.8660254, 0.8660254, 0.5],
                    [-1.5, 0.8660254, -0.8660254, 0.5],
                    [-0.0, -1.73205081, -0.0, -1.0],
                    [1.5, -0.8660254, -0.8660254, 0.5],
                    [1.5, -0.8660254, -0.8660254, 0.5],
                    [0.0, 1.73205081, -0.0, -1.0],
                    [-0.0, -0.0, -1.73205081, -1.0],
                    [-0.0, -0.0, -1.73205081, -1.0],
                    [0.0, 0.0, 0.0, 2.0],
                ]
            ),
        )

    def test_modelmatrix_verbosity(self):
        conf_design = oapackage.exampleArray(41, 0)
        with mock.patch("sys.stdout", new_callable=io.StringIO) as mock_stdout:
            oapackage.array2modelmatrix(conf_design, "i", 1)
            stdout = mock_stdout.getvalue()
            self.assertIn("array2modelmatrix: type conference, model_type_idx 2", stdout)


class TestArrayLink(unittest.TestCase):
    def test_selectFirstColumns(self):
        al = oapackage.exampleArray(41, 0)
        al = al.selectFirstColumns(3)
        self.assertEqual(al.n_columns, 3)

        al = oapackage.array_link()
        with self.assertRaises(RuntimeError):
            al = al.selectFirstColumns(1)

    def test_strength(self):
        example_strength_pairs = [(2, 2), (3, 3), (4, 2), (6, 2), (7, 2)]
        for idx, strength in example_strength_pairs:
            array = oapackage.exampleArray(idx, 0)
            self.assertEqual(array.strength(), strength)

    def test_basic_array_link_functionality(self):
        al2a = oapackage.array_link(2, 2, 0)
        al2b = oapackage.array_link(2, 2, 0)
        al2b.setconstant(1)

        al3 = oapackage.array_link(3, 2, 0)
        self.assertTrue(al2a != al2b)
        self.assertFalse(al2a == al2b)
        self.assertTrue(al2a.equalsize(al2b))
        self.assertFalse(al2a.equalsize(al3))
        self.assertTrue(al2a < al2b)
        self.assertFalse(al2b < al2a)

    def test_array_link_dimensions(self):
        for example_idx in [0, 2, 6, 10]:
            al = oapackage.exampleArray(example_idx, 0)
            self.assertEqual(al.size, al.n_rows * al.n_columns)
            self.assertEqual(al.shape, (al.n_rows, al.n_columns))
        with self.assertRaises(AttributeError):
            al.shape = 1

    def test_is_ortogonal_array(self):
        al = oapackage.exampleArray(1, 0)
        self.assertTrue(al.is_orthogonal_array())
        al[1, 2] = -1
        self.assertFalse(al.is_orthogonal_array())
        al = oapackage.exampleArray(41, 0)
        self.assertFalse(al.is_orthogonal_array())

    def test_array_class_functions(self):
        al = oapackage.exampleArray(1, 0)
        self.assertTrue(al.is2level())
        self.assertFalse(al.is_mixed_level())

        al = oapackage.exampleArray(11, 0)
        self.assertTrue(al.is2level())
        self.assertFalse(al.is_mixed_level())

        al = oapackage.array_link(2, 2, 0)
        al[1, 1] = -1
        self.assertFalse(al.is2level())
        self.assertFalse(al.is_mixed_level())


class TestArraydata_t(unittest.TestCase):
    def test_factor_levels(self):
        array = oapackage.exampleArray(5, 0)
        arrayclass = oapackage.arraylink2arraydata(array)
        factor_levels = arrayclass.factor_levels()
        self.assertEqual(factor_levels, (4, 3, 2, 2, 2))

    def test_arraydata_t_oaindex(self):
        for ii in range(1, 4):
            arrayclass = oapackage.arraydata_t([2, 2, 2], 4 * ii, 2, 3)
            self.assertEqual(arrayclass.oaindex, ii)

        expected_regex = ".*warning: no orthogonal arrays exist with the specified strength.*"
        with self.assertWarnsRegex(RuntimeWarning, expected_regex):
            arrayclass = oapackage.arraydata_t([4, 3, 3], 20, 2, 3)
        self.assertEqual(arrayclass.oaindex, 0)

        with self.assertWarnsRegex(RuntimeWarning, expected_regex):
            arrayclass = oapackage.arraydata_t([2, 3, 4], 20, 2, 3)

        with self.assertWarnsRegex(RuntimeWarning, expected_regex):
            arrayclass = oapackage.arraydata_t([6, 5], 10, 1, 2)
            self.assertEqual(arrayclass.oaindex, 0)


class TestJcharacteristics(unittest.TestCase):
    def test_jstruct_conference(self):
        conf_design = oapackage.exampleArray(30, 0)
        js = oapackage.jstructconference_t(conf_design, 4)
        self.assertEqual(js.Jvalues(), (4, 0))

        with self.assertRaises(RuntimeError):
            js = oapackage.jstructconference_t(conf_design, 3)

        array = oapackage.array_link(10, 4, 0)
        with mock.patch("sys.stdout", new_callable=io.StringIO) as mock_stdout:
            with self.assertRaises(Exception):
                js = oapackage.jstructconference_t(array, 4)
            std_output = mock_stdout.getvalue()
            self.assertIn("array is not conference design", std_output)

    def test_Jcharacteristics(self):
        al = oapackage.exampleArray(30, 0)
        jx = al.Jcharacteristics(4)
        self.assertEqual(jx, (0,))
        jx = al.Jcharacteristics(2)
        self.assertEqual(jx, (0, 0, 0, 0, 0, 0))

        al = oapackage.exampleArray(48, 0).selectFirstColumns(6)
        jx = al.Jcharacteristics(4)
        self.assertEqual(jx, (4, 4, 4, 0, 8, -8, 8, 4, -4, -4, 8, 4, -4, 8, -4))


class TestConferenceDesigns(unittest.TestCase):
    def test_conf2dsd(self):
        al = oapackage.exampleArray(42, 0)
        dsd = oapackage.conference2DSD(al)
        self.assertEqual(dsd.n_rows, 2 * al.n_rows + 1)

        conf = np.array(al)
        dsd = np.array(dsd)

        np.testing.assert_array_equal(conf, dsd[0 : al.n_rows, :])
        np.testing.assert_array_equal(conf, -dsd[al.n_rows : 2 * al.n_rows, :])
        np.testing.assert_array_equal(0 * conf[0], dsd[-1, :])

    def test_isConferenceFoldover(self):
        al = oapackage.exampleArray(37, 0)
        self.assertTrue(oapackage.isConferenceFoldover(al))
        al[2, 0] = -1
        self.assertFalse(oapackage.isConferenceFoldover(al))
        al = oapackage.exampleArray(7, 0)
        self.assertFalse(oapackage.isConferenceFoldover(al))
        al = oapackage.exampleArray(45, 0)
        self.assertFalse(oapackage.isConferenceFoldover(al))

    def test_double_conference_foldover_permutation(self):
        al = oapackage.exampleArray(37, 0)
        expected = [0, 2, 3, 4, 5, 6, 7, 8, 9, 10, 1, 11, 16, 19, 17, 18, 15, 14, 13, 12]
        permutation = oapackage.double_conference_foldover_permutation(al)
        self.assertEqual(list(permutation), expected)

        N = int(al.n_rows / 2)
        folded = np.array(oapackage.array_link(np.array(al)[permutation, :]))
        np.testing.assert_array_equal(folded[0:N, :], -folded[N:, :])

        expected[5], expected[3] = expected[3], expected[5]
        expected[-1], expected[4] = expected[4], expected[-1]
        expected[-3], expected[7] = expected[7], expected[-3]
        al = oapackage.array_link(np.array(al)[expected, :])
        permutation = oapackage.double_conference_foldover_permutation(al)
        folded = np.array(oapackage.array_link(np.array(al)[permutation, :]))
        N = int(al.n_rows / 2)
        np.testing.assert_array_equal(folded[0:N, :], -folded[N:, :])

        al = oapackage.exampleArray(45, 0)
        permutation = oapackage.double_conference_foldover_permutation(al)
        self.assertEqual(permutation[0], -1)

        al = oapackage.exampleArray(5, 0)
        with self.assertRaises(RuntimeError):
            permutation = oapackage.double_conference_foldover_permutation(al)

        al = oapackage.array_link(3, 4, 0)
        with self.assertRaises(RuntimeError):
            permutation = oapackage.double_conference_foldover_permutation(al)


class TestMathTools(unittest.TestCase):
    def test_krawtchouk(self):
        self.assertEqual(oapackage.krawtchouk(6, 3, 2, 2), 0)
        self.assertEqual(oapackage.krawtchouk(4, 0, 5, 3), 80)
        self.assertEqual(oapackage.krawtchouk(6, 0, 2, 2), 0)
        self.assertEqual(oapackage.krawtchouk(6, 5, 2, 3), 0)
        self.assertEqual(oapackage.krawtchouk(1, 1, 1, 1), -1)
        self.assertEqual(oapackage.krawtchouk(14, 0, 15, 2), 15)


class TestArrayFiles(unittest.TestCase):
    def test_write_latex_format(self):
        lst = [oapackage.exampleArray(2, 0)]
        filename = tempfile.mktemp(suffix=".tex")
        oapackage.writearrayfile(filename, oapackage.arraylist_t(lst), oapackage.ALATEX)
        with open(filename) as fid:
            latex = fid.read()
        self.assertIsInstance(latex, str)


class TestCppLibrary(unittest.TestCase):
    def setUp(self):
        if getattr(self, "assertRaisesRegex", None) is None:
            self.assertRaisesRegex = self.assertRaisesRegexp

    def test_get_time_ms(self):
        t0 = oapackage.get_time_ms()
        time.sleep(0.2)
        dt = oapackage.get_time_ms(t0)
        np.testing.assert_almost_equal(dt, 0.2, decimal=1)

    def test_root_form(self):
        array = oapackage.exampleArray(1, 0)
        self.assertTrue(oapackage.is_root_form(array, 2))
        self.assertFalse(oapackage.is_root_form(array, 5))

    def test_runExtend_increasing_factor_levels(self):
        """We test the usage of an increasing factor levels raises are warning, but nevertheless gives the correct results"""
        N = 18
        k = 9
        t = 2
        l = [2, 3]
        rr = []

        expected_regex = (
            ".*warning: the factor levels of the structure are not sorted, this can lead to undefined behaviour.*"
        )
        with self.assertWarnsRegex(RuntimeWarning, expected_regex):
            oapackage.oahelper.runExtend(N, k, t, l, verbose=1, nums=rr)

        self.assertEqual(rr, [3, 15, 48, 19, 12, 3, 0])

    def test_print_column(self):
        with redirect_stdout(io.StringIO()) as f:
            oapackage.print_column([1, 2, 3])
        self.assertEqual(f.getvalue(), "  1  2  3\n")

    def test_projectionDOFvalues(self):
        array = oapackage.exampleArray(5, 0)
        arrayclass = oapackage.arraylink2arraydata(array)
        dof_values = oapackage.projectionDOFvalues(array)
        sg = oapackage.symmetry_group(dof_values, False)
        with mock.patch("sys.stdout", new_callable=io.StringIO) as mock_stdout:
            sg.show(1)
        self.assertIn("symmetry group: 5 elements, 5 subgroups: 1 1 1 1 1", mock_stdout.getvalue())

        for column in range(array.n_columns):
            dof_element = list(dof_values[column].raw_values())
            self.assertEqual(dof_element[0], -arrayclass.factor_levels()[column])

    def test_exampleArray(self):
        with mock.patch("sys.stdout", new_callable=io.StringIO) as mock_stdout:
            al = oapackage.exampleArray(5, 1)
            self.assertEqual(mock_stdout.getvalue(), "exampleArray 5: array 0 in OA(24, 2, 4 3 2^a)\n")
            self.assertEqual(al.md5(), "3885c883d3bee0c7546511255bb5c3ae")

        with mock.patch("sys.stdout", new_callable=io.StringIO) as mock_stdout:
            al = oapackage.exampleArray(-1, 1)
            lines = mock_stdout.getvalue().strip().split("\n")
            self.assertTrue(np.all([lines[ii].startswith(f"exampleArray {ii}:") for ii in range(len(lines))]))

        al = oapackage.exampleArray(51, 0)
        self.assertEqual(al.md5(), "662c1d9c51475b42539620385fa22338")

        with self.assertRaises(Exception):
            al = oapackage.exampleArray(1000, 0)

    def test_mvalue_t(self):
        input_vector = [1.0, 2.0, 2.0]
        m = oapackage.mvalue_t_double(input_vector)
        self.assertEqual(m.size(), len(input_vector))
        self.assertEqual(list(m.values), input_vector)
        self.assertIsInstance(m.__array_interface__, dict)

        input_vector = [1, 2, -1]
        mvalue_long = oapackage.mvalue_t_long(input_vector)
        self.assertEqual(list(mvalue_long.values), input_vector)

    def test_splits(self):
        self.assertEqual(oapackage.splitTag([10, 12]), "10.12")
        self.assertEqual(oapackage.splitFile([]), "")
        self.assertEqual(oapackage.splitDir([1, 2]), "sp0-split-1" + os.path.sep + "sp1-split-2" + os.path.sep)

    def test_array_transformation_t(self):
        at = oapackage.array_transformation_t()
        with patch("sys.stdout", new_callable=io.StringIO) as mock_stdout:
            at.show()
            std_output = mock_stdout.getvalue()
            self.assertEqual(std_output, "array transformation: no class defined\n")

        al = oapackage.exampleArray(0, 0)
        arrayclass = oapackage.arraylink2arraydata(al)
        at = oapackage.array_transformation_t(arrayclass)
        at.setcolperm([1, 0])
        with patch("sys.stdout", new_callable=io.StringIO) as mock_stdout:
            _ = at.show()
            std_output = mock_stdout.getvalue()
            self.assertEqual(
                std_output,
                "array transformation: N 8\ncolumn permutation: {1,0}\nlevel perms:\n{0,1}\n{0,1}\nrow permutation: {0,1,2,3,4,5,6,7}\n",
            )
        self.assertEqual(at.colperm(), (1, 0))

    def test_arraylink2arraydata(self):
        al = oapackage.exampleArray(0, 0)
        adata = oapackage.arraylink2arraydata(al)
        self.assertEqual(str(adata), r"arrayclass: N 8, k 2, strength 2, s {2,2}, order 0")
        al = oapackage.array_link(4, 4, 0)
        al.setconstant(0)
        adata = oapackage.arraylink2arraydata(al)
        al.setconstant(-1)
        with self.assertRaises(RuntimeError):
            _ = oapackage.arraylink2arraydata(al)

        for ii in [0, 4, 8, 10]:
            al = oapackage.exampleArray(ii, 0)
            arrayclass = oapackage.arraylink2arraydata(al, strength=-1)
            self.assertEqual(arrayclass.strength, al.strength())

    def test_exception_handling(self):
        with self.assertRaises(RuntimeError):
            oapackage.mycheck_handler("file", "function", 10, 0, "hi")
        oapackage.mycheck_handler("file", "function", 10, 1, "hi")

        with self.assertRaises(RuntimeError):
            oapackage.throw_runtime_exception("should throw error")

        al = oapackage.oalib.exampleArray(18, 0)
        with self.assertRaisesRegex(RuntimeError, "array cannot have negative elements"):
            _ = oapackage.array2eigenModelMatrixMixed(al, 2)

    def test_array2eigenModelMatrixMixed(self):
        array = oapackage.exampleArray(5, 0)
        model = oapackage.array2eigenModelMatrixMixed(array)
        expected_main_effects = np.frombuffer(
            b"\xcc;\x7ff\x9e\xa0\xf6\xbf=,\x0cp\xbd \xea\xbf\x1c3\x90E\xa7y\xe2\xbf.!\t\x14\x8e\x98\xf3\xbf\xcc;\x7ff\x9e\xa0\xe6\xbf\x00\x00\x00\x00\x00\x00\xf0\xbf\x00\x00\x00\x00\x00\x00\xf0\xbf\x00\x00\x00\x00\x00\x00\xf0\xbf\xcc;\x7ff\x9e\xa0\xf6\xbf=,\x0cp\xbd \xea\xbf\x1c3\x90E\xa7y\xe2\xbf.!\t\x14\x8e\x98\xf3\xbf\xcc;\x7ff\x9e\xa0\xe6\xbf\x00\x00\x00\x00\x00\x00\xf0\xbf\x00\x00\x00\x00\x00\x00\xf0\xbf\x00\x00\x00\x00\x00\x00\xf0\xbf\xcc;\x7ff\x9e\xa0\xf6\xbf=,\x0cp\xbd \xea\xbf\x1c3\x90E\xa7y\xe2\xbf.!\t\x14\x8e\x98\xf3?\xcc;\x7ff\x9e\xa0\xe6\xbf\x00\x00\x00\x00\x00\x00\xf0\xbf\x00\x00\x00\x00\x00\x00\xf0\xbf\x00\x00\x00\x00\x00\x00\xf0\xbf\xcc;\x7ff\x9e\xa0\xf6\xbf=,\x0cp\xbd \xea\xbf\x1c3\x90E\xa7y\xe2\xbf.!\t\x14\x8e\x98\xf3?\xcc;\x7ff\x9e\xa0\xe6\xbf\x00\x00\x00\x00\x00\x00\xf0?\x00\x00\x00\x00\x00\x00\xf0?\x00\x00\x00\x00\x00\x00\xf0?\xcc;\x7ff\x9e\xa0\xf6\xbf=,\x0cp\xbd \xea\xbf\x1c3\x90E\xa7y\xe2\xbf\x00\x00\x00\x00\x00\x00\x00\x00\xcc;\x7ff\x9e\xa0\xf6?\x00\x00\x00\x00\x00\x00\xf0?\x00\x00\x00\x00\x00\x00\xf0?\x00\x00\x00\x00\x00\x00\xf0?\xcc;\x7ff\x9e\xa0\xf6\xbf=,\x0cp\xbd \xea\xbf\x1c3\x90E\xa7y\xe2\xbf\x00\x00\x00\x00\x00\x00\x00\x00\xcc;\x7ff\x9e\xa0\xf6?\x00\x00\x00\x00\x00\x00\xf0?\x00\x00\x00\x00\x00\x00\xf0?\x00\x00\x00\x00\x00\x00\xf0?\xcc;\x7ff\x9e\xa0\xf6?=,\x0cp\xbd \xea\xbf\x1c3\x90E\xa7y\xe2\xbf.!\t\x14\x8e\x98\xf3\xbf\xcc;\x7ff\x9e\xa0\xe6\xbf\x00\x00\x00\x00\x00\x00\xf0\xbf\x00\x00\x00\x00\x00\x00\xf0?\x00\x00\x00\x00\x00\x00\xf0?\xcc;\x7ff\x9e\xa0\xf6?=,\x0cp\xbd \xea\xbf\x1c3\x90E\xa7y\xe2\xbf.!\t\x14\x8e\x98\xf3\xbf\xcc;\x7ff\x9e\xa0\xe6\xbf\x00\x00\x00\x00\x00\x00\xf0\xbf\x00\x00\x00\x00\x00\x00\xf0?\x00\x00\x00\x00\x00\x00\xf0?\xcc;\x7ff\x9e\xa0\xf6?=,\x0cp\xbd \xea\xbf\x1c3\x90E\xa7y\xe2\xbf.!\t\x14\x8e\x98\xf3?\xcc;\x7ff\x9e\xa0\xe6\xbf\x00\x00\x00\x00\x00\x00\xf0\xbf\x00\x00\x00\x00\x00\x00\xf0\xbf\x00\x00\x00\x00\x00\x00\xf0?\xcc;\x7ff\x9e\xa0\xf6?=,\x0cp\xbd \xea\xbf\x1c3\x90E\xa7y\xe2\xbf.!\t\x14\x8e\x98\xf3?\xcc;\x7ff\x9e\xa0\xe6\xbf\x00\x00\x00\x00\x00\x00\xf0?\x00\x00\x00\x00\x00\x00\xf0?\x00\x00\x00\x00\x00\x00\xf0\xbf\xcc;\x7ff\x9e\xa0\xf6?=,\x0cp\xbd \xea\xbf\x1c3\x90E\xa7y\xe2\xbf\x00\x00\x00\x00\x00\x00\x00\x00\xcc;\x7ff\x9e\xa0\xf6?\x00\x00\x00\x00\x00\x00\xf0?\x00\x00\x00\x00\x00\x00\xf0\xbf\x00\x00\x00\x00\x00\x00\xf0\xbf\xcc;\x7ff\x9e\xa0\xf6?=,\x0cp\xbd \xea\xbf\x1c3\x90E\xa7y\xe2\xbf\x00\x00\x00\x00\x00\x00\x00\x00\xcc;\x7ff\x9e\xa0\xf6?\x00\x00\x00\x00\x00\x00\xf0?\x00\x00\x00\x00\x00\x00\xf0\xbf\x00\x00\x00\x00\x00\x00\xf0\xbf\x00\x00\x00\x00\x00\x00\x00\x00=,\x0cp\xbd \xfa?\x1c3\x90E\xa7y\xe2\xbf.!\t\x14\x8e\x98\xf3\xbf\xcc;\x7ff\x9e\xa0\xe6\xbf\x00\x00\x00\x00\x00\x00\xf0?\x00\x00\x00\x00\x00\x00\xf0\xbf\x00\x00\x00\x00\x00\x00\xf0?\x00\x00\x00\x00\x00\x00\x00\x00=,\x0cp\xbd \xfa?\x1c3\x90E\xa7y\xe2\xbf.!\t\x14\x8e\x98\xf3\xbf\xcc;\x7ff\x9e\xa0\xe6\xbf\x00\x00\x00\x00\x00\x00\xf0?\x00\x00\x00\x00\x00\x00\xf0\xbf\x00\x00\x00\x00\x00\x00\xf0?\x00\x00\x00\x00\x00\x00\x00\x00=,\x0cp\xbd \xfa?\x1c3\x90E\xa7y\xe2\xbf.!\t\x14\x8e\x98\xf3?\xcc;\x7ff\x9e\xa0\xe6\xbf\x00\x00\x00\x00\x00\x00\xf0\xbf\x00\x00\x00\x00\x00\x00\xf0?\x00\x00\x00\x00\x00\x00\xf0\xbf\x00\x00\x00\x00\x00\x00\x00\x00=,\x0cp\xbd \xfa?\x1c3\x90E\xa7y\xe2\xbf.!\t\x14\x8e\x98\xf3?\xcc;\x7ff\x9e\xa0\xe6\xbf\x00\x00\x00\x00\x00\x00\xf0?\x00\x00\x00\x00\x00\x00\xf0\xbf\x00\x00\x00\x00\x00\x00\xf0?\x00\x00\x00\x00\x00\x00\x00\x00=,\x0cp\xbd \xfa?\x1c3\x90E\xa7y\xe2\xbf\x00\x00\x00\x00\x00\x00\x00\x00\xcc;\x7ff\x9e\xa0\xf6?\x00\x00\x00\x00\x00\x00\xf0\xbf\x00\x00\x00\x00\x00\x00\xf0?\x00\x00\x00\x00\x00\x00\xf0\xbf\x00\x00\x00\x00\x00\x00\x00\x00=,\x0cp\xbd \xfa?\x1c3\x90E\xa7y\xe2\xbf\x00\x00\x00\x00\x00\x00\x00\x00\xcc;\x7ff\x9e\xa0\xf6?\x00\x00\x00\x00\x00\x00\xf0\xbf\x00\x00\x00\x00\x00\x00\xf0?\x00\x00\x00\x00\x00\x00\xf0\xbf\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xaaLX\xe8z\xb6\xfb?.!\t\x14\x8e\x98\xf3\xbf\xcc;\x7ff\x9e\xa0\xe6\xbf\x00\x00\x00\x00\x00\x00\xf0?\x00\x00\x00\x00\x00\x00\xf0?\x00\x00\x00\x00\x00\x00\xf0\xbf\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xaaLX\xe8z\xb6\xfb?.!\t\x14\x8e\x98\xf3\xbf\xcc;\x7ff\x9e\xa0\xe6\xbf\x00\x00\x00\x00\x00\x00\xf0?\x00\x00\x00\x00\x00\x00\xf0?\x00\x00\x00\x00\x00\x00\xf0\xbf\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xaaLX\xe8z\xb6\xfb?.!\t\x14\x8e\x98\xf3?\xcc;\x7ff\x9e\xa0\xe6\xbf\x00\x00\x00\x00\x00\x00\xf0\xbf\x00\x00\x00\x00\x00\x00\xf0?\x00\x00\x00\x00\x00\x00\xf0?\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xaaLX\xe8z\xb6\xfb?.!\t\x14\x8e\x98\xf3?\xcc;\x7ff\x9e\xa0\xe6\xbf\x00\x00\x00\x00\x00\x00\xf0?\x00\x00\x00\x00\x00\x00\xf0\xbf\x00\x00\x00\x00\x00\x00\xf0\xbf\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xaaLX\xe8z\xb6\xfb?\x00\x00\x00\x00\x00\x00\x00\x00\xcc;\x7ff\x9e\xa0\xf6?\x00\x00\x00\x00\x00\x00\xf0\xbf\x00\x00\x00\x00\x00\x00\xf0\xbf\x00\x00\x00\x00\x00\x00\xf0?\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xaaLX\xe8z\xb6\xfb?\x00\x00\x00\x00\x00\x00\x00\x00\xcc;\x7ff\x9e\xa0\xf6?\x00\x00\x00\x00\x00\x00\xf0\xbf\x00\x00\x00\x00\x00\x00\xf0\xbf\x00\x00\x00\x00\x00\x00\xf0?"
        ).reshape(24, 8)
        main_effects, interaction_effects = model[0], model[1]
        np.testing.assert_array_equal(main_effects, expected_main_effects)
        self.assertEqual(interaction_effects.shape, (24, 24))

    def test_mycheck_handler(self):
        oapackage.mycheck_handler("a", "b", 1, 1, "test mycheck_handler")
        with self.assertRaises(RuntimeError):
            oapackage.mycheck_handler("a", "b", 1, 0, "mycheck_handler raise")

    def test_arrayrankInfo(self):
        with patch("sys.stdout", new_callable=io.StringIO) as mock_stdout:
            rank = oapackage.arrayrankInfo(oapackage.exampleArray(21))
            self.assertEqual(rank, 4)
            self.assertIn("FullPivLU: rank 4", mock_stdout.getvalue())

    def test_rankStructure(self):
        array = oapackage.exampleArray(45, 0)
        array2 = oapackage.exampleArray(46, 0)
        rank_structure = oapackage.rankStructure(array)
        rank_structure.verbose = 2
        with patch("sys.stdout", new_callable=io.StringIO):
            rank_structure.info()

        rank_structure.verbose = 0
        rank_direct = rank_structure.rankxf(array)
        rank = rank_structure.rankxf(array)
        rank2 = rank_structure.rankxf(array2)
        self.assertEqual(rank_direct, 20)
        self.assertEqual(rank, 20)
        self.assertEqual(rank2, 20)

    def test_distance_distribution_sum(self):
        array = oapackage.exampleArray(1, 1)
        array = array.selectFirstColumns(3)

        D = oapackage.distance_distribution_mixed(array, 0)
        np.testing.assert_array_equal(D, [2.5, 4.5, 7.5, 1.5])
        self.assertEqual(np.sum(D), 16.0)

    def distance_distribution(self):
        al = oapackage.array_link(2, 2, 0)
        distance_distrib = oapackage.distance_distribution(al)
        self.assertEqual(distance_distrib, (2.0, 0.0, 0.0))

        al[1, 0] = 1
        distance_distrib = oapackage.distance_distribution(al)
        self.assertEqual(distance_distrib, (1.0, 1.0, 0.0))

        al = oapackage.exampleArray(2, 0)

        distance_distrib = oapackage.distance_distribution(al)
        self.assertEqual(distance_distrib, (1.25, 0.75, 1.5, 6.5, 5.25, 0.75, 0.0))

    def test_distance_distribution_shape(self):
        array = oapackage.exampleArray(1, 0)
        dims = oapackage.distance_distribution_shape(oapackage.arraylink2arraydata(array))
        self.assertEqual(dims, (6,))

    def test_distance_distribution_mixed(self):
        array = oapackage.exampleArray(1, 0).selectFirstColumns(3)
        dims = oapackage.distance_distribution_shape(oapackage.arraylink2arraydata(array))

        B0 = oapackage.distance_distribution_mixed(array)
        dd0 = np.array(B0)

        B = oapackage.ndarray_double(dims)
        oapackage.distance_distribution_mixed_inplace(array, B, verbose=0)
        dd_mixed = np.asarray(B)

        dd = oapackage.distance_distribution(array)

        np.testing.assert_array_equal(dd0, dd)
        np.testing.assert_array_equal(dd0, dd_mixed)

    def test_distance_distribution_mixed_2(self):
        A = oapackage.exampleArray(5, 1)
        N = A.n_rows
        with mock.patch("sys.stdout", new_callable=io.StringIO) as mock_stdout:
            d = oapackage.distance_distribution_mixed(A, verbose=1)
            s = mock_stdout.getvalue()
        self.assertTrue(s.startswith("distance_distribution_mixed: \nB[0] = B[0][0][0] = 1.666667"))
        dd = np.array(d)

        gt = np.array(
            [[[40.0, 0.0, 0.0, 8.0], [16.0, 16.0, 16.0, 48.0]], [[0.0, 24.0, 120.0, 0.0], [16.0, 176.0, 80.0, 16.0]]]
        )
        self.assertEqual(dd.shape, tuple(d.dims))
        np.testing.assert_array_equal(N * dd, gt)

    def test_Defficiencies(self):
        array = oapackage.exampleArray(0, 0)
        efficiencies = array.Defficiencies()
        self.assertEqual(efficiencies, (1.0, 1.0, 1.0))

        self.assertAlmostEqual(array.Eefficiency(), 1)
        self.assertAlmostEqual(array.VIFefficiency(), 1)

        array = oapackage.exampleArray(11, 0)
        efficiencies = array.Defficiencies()
        self.assertEqual(efficiencies, (0.8879176205539137, 0.8058954593581515, 0.9936928808878267))

        self.assertAlmostEqual(array.Eefficiency(), 0.3602369388406959)
        self.assertAlmostEqual(array.VIFefficiency(), 1.2648199507220503)

        for idx in [0, 4, 11, 16]:
            array = oapackage.exampleArray(idx, 0)
            efficiencies = array.Defficiencies()
            D = array.Defficiency()
            Ds = array.DsEfficiency()
            self.assertAlmostEqual(efficiencies[0], D)
            self.assertAlmostEqual(efficiencies[1], Ds)

    def test_projection_efficiencies(self):
        al = oapackage.exampleArray(11, 0)
        with mock.patch("sys.stdout", new_callable=io.StringIO) as mock_stdout:
            d = oapackage.projDeff(al, 3, 1)
            D = al.selectFirstColumns(3).Defficiency()
            std_output = mock_stdout.getvalue()
            self.assertIn("projDeff: k 8, kp 3: start with 56 combinations", std_output)
        self.assertEqual(D, d[0])
        np.testing.assert_almost_equal(np.mean(d), 0.99064112542249538329031111061340197921)

        pec_seq = oapackage.PECsequence(al)
        np.testing.assert_equal(pec_seq, (1.0,) * len(pec_seq))
        pic_seq = oapackage.PICsequence(al)
        np.testing.assert_almost_equal(
            pic_seq,
            (
                0.9985780064264659,
                0.9965697009006985,
                0.9906411254224957,
                0.9797170906488152,
                0.9635206782887167,
                0.9421350381959234,
                0.9162739059686846,
                0.8879176205539139,
            ),
        )

    def test_PICsequence_length(self):
        array = oapackage.exampleArray(8, 0)
        pic = oapackage.PICsequence(array)
        self.assertEqual(len(pic), array.n_columns)

    def test_arraylink(self):
        al = oapackage.exampleArray(0, 0)
        self.assertEqual(al.at(0, 1), 0)
        self.assertEqual(al.at(0), 0)
        with self.assertRaises(IndexError):
            al.at(-1, 0)
        with self.assertRaises(IndexError):
            al.at(-1)
        with self.assertRaises(IndexError):
            al.at(0, al.n_columns)

    def test_arraylink_slicing(self):
        numpy_array = np.arange(0, 6 * 10).reshape((6, 10))

        al = oapackage.makearraylink(numpy_array)
        self.assertTrue(al[0] == numpy_array.flatten()[0])
        self.assertTrue(al[0, 1] == numpy_array[0, 1])
        self.assertTrue(al[4, 2] == numpy_array[4, 2])
        np.testing.assert_equal(al[0:4, 1:5], np.array(al)[0:4, 1:5])
        np.testing.assert_equal(al[0:1, 0:10:2], np.array(al)[0:1, 0:10:2])
        np.testing.assert_equal(al[3, 3::], np.array(al)[3:4, 3::])
        np.testing.assert_equal(al[2, :8:2], np.array(al)[2:3, :8:2])
        np.testing.assert_equal(al[2:3, :8:2], np.array(al)[2:3, :8:2])

        with self.assertRaises(IndexError):
            _ = al[-1, 1]

    def test_conference_generation(self):
        al = oapackage.exampleArray(42, 0)
        lst = [al]
        conference_type = oapackage.conference_t(al.n_rows, al.n_rows, 0)

        with mock.patch("sys.stdout", new_callable=io.StringIO) as mock_stdout:
            extensions = oapackage.extend_conference_restricted(lst, conference_type, verbose=1)
            std_output = mock_stdout.getvalue()
            self.assertIn("extend_conference", std_output)
        self.assertEqual(len(extensions), 10)

        self.assertEqual(extensions[0].md5(), "f759e75d3ce6adda5489fed4c528a6fb")
        self.assertEqual(extensions[-1].md5(), "ec13ed02c70dbc9975a99e70e23154b3")


if __name__ == "__main__":
    """ Test code """
    unittest.main()



"""Module to generate D-optimal designs

For more information see: https://doi.org/10.1080/00401706.2016.1142903

Pieter Eendebak <pieter.eendebak@gmail.com>

"""

import logging
import os
import time
import warnings
from collections.abc import Callable
from typing import Any

import matplotlib
import matplotlib.cm
import matplotlib.pyplot as plt
import numpy as np
import numpy.typing
import oalib

import oapackage.markup as markup
import oapackage.oahelper as oahelper
from oapackage.markup import oneliner as e

FloatArray = np.typing.NDArray[np.float64]

# %%


def array2Dtable(array_list: list, verbose: int = 1, titlestr: str | None = None):
    """Generate HTML table with information about for a list of designs

    Args:
        array_list: list of arrays
        verbose: verbosity level
        titlestr: Not used
    """
    page = markup.page()
    page.table(style=" border-collapse: collapse;")
    page.tr(style="font-weight: bold; border-bottom: solid 1px black;")
    page.th("Array", style="padding-right:30px; ")
    page.th(("D-efficiency", "Ds-efficiency", "D1-efficiency"), style="padding-right:14px;")
    page.th(("GWLP"), style="padding-right:14px;")
    page.tr.close()
    for ii, al in enumerate(array_list):
        aidx = ii
        (D, Ds, D1) = al.Defficiencies()
        gwlp = al.GWLP()
        page.tr(style="font-weight: normal;")
        page.td("%d" % aidx, style="padding-right:10px;")
        for statistic in [D, Ds, D1]:
            page.td(f"{statistic:.4f}", style="padding-right:1px;")
        gstr = oahelper.gwlp2str(gwlp)
        page.td(e.small(gstr), style="padding-right:1px;")
        page.tr.close()
    page.table.close()
    return page


# %%


def generateDscatter(
    dds,
    second_index=0,
    first_index=1,
    lbls=None,
    ndata=3,
    nofig=False,
    fig=20,
    scatterarea=80,
    verbose=0,
    setWindowRectangle=False,
):
    """Generate scatter plot for D and Ds efficiencies

    Args:
        dds (array): array with D-efficiencies
    Returns:
        dict: contains handles to plotting elements
    """
    data = dds.T
    pp = oahelper.createPareto(dds)
    paretoidx = np.array(pp.allindices())

    nn = dds.shape[0]
    area = (
        scatterarea
        * np.ones(
            nn,
        )
        / 2
    )
    area[np.array(pp.allindices())] = scatterarea
    alpha = 1.0

    if dds.shape[1] > ndata:
        colors = dds[:, ndata]
    else:
        colors = np.zeros((nn, 1))

    idx = np.unique(colors).astype(int)

    if verbose:
        print(f"generateDscatter: unique colors: {idx}")
    ncolors = idx.size
    try:
        import brewer2mpl  # ty: ignore

        ncolors = max(ncolors, 4)
        mycmap = brewer2mpl.get_map("Set1", "qualitative", ncolors).mpl_colors
    except BaseException:
        mycmap = [matplotlib.cm.jet(ii) for ii in np.linspace(0, 1, ncolors)]

    nonparetoidx = np.setdiff1d(range(nn), paretoidx)
    if lbls is None:
        lbls = ["%d" % i for i in range(len(idx))]

    if fig is not None:
        figh = plt.figure(fig)
        plt.clf()
        figh.set_facecolor("w")
        ax = plt.subplot(111)

        ax.scatter(
            data[first_index, nonparetoidx],
            data[second_index, nonparetoidx],
            s=0.33 * scatterarea,
            c=[(0.5, 0.5, 0.5)],
            linewidths=0,
            alpha=alpha,
            label="Non-pareto design",
        )

        for jj, ii in enumerate(idx):
            gidx = (colors == ii).nonzero()[0]
            gp = np.intersect1d(paretoidx, gidx)

            color = mycmap[jj]
            cc = [color] * len(gp)
            if verbose:
                print("index %d: %d points" % (ii, gidx.size))
            ax.scatter(
                data[first_index, gp],
                data[second_index, gp],
                s=scatterarea,
                c=cc,
                linewidths=0,
                alpha=alpha,
                label=lbls[jj],
            )
            plt.draw()

        if data[second_index, :].std() < 1e-3:
            y_formatter = matplotlib.ticker.ScalarFormatter(useOffset=False)
            ax.yaxis.set_major_formatter(y_formatter)

        xlabelhandle = plt.xlabel("$D_s$-efficiency", fontsize=16)
        plt.ylabel("D-efficiency", fontsize=16)

        if setWindowRectangle:
            try:
                oahelper.setWindowRectangle(10, 10, 860, 600)
            except Exception as ex:
                print("generateDscatter: setWindowRectangle failed")
                logging.exception(ex)

        plt.axis("image")
        pltlegend = ax.legend(loc=3, scatterpoints=1)  # , fontcolor=almost_black)
        if not nofig:
            plt.show()
        ax.grid(visible=True, which="both", color="0.85", linestyle="-")
        ax.set_axisbelow(True)

        if nofig:
            plt.close(figh.number)
        else:
            plt.draw()
            plt.pause(1e-3)
    else:
        ax = None
        xlabelhandle = None
        pltlegend = None
    hh = dict({"ax": ax, "xlabelhandle": xlabelhandle, "pltlegend": pltlegend})
    return hh


def generateDpage(
    outputdir,
    arrayclass,
    dds,
    allarrays,
    fig=20,
    optimfunc=(1, 0, 0),
    nofig=False,
    urlprefix="",
    makeheader=True,
    verbose=1,
    lbls=None,
):
    """Helper function to generate web page with D-optimal design results"""
    if verbose:
        print(f"generateDpage: dds {str(dds.shape)}")

    pp = oahelper.createPareto(dds)

    narrays = dds.shape[0]
    npareto = pp.number()

    if verbose:
        print("generateDpage: narrays %d" % narrays)

    xstr = oahelper.series2htmlstr(arrayclass, case=1)
    xstrplain = oahelper.series2htmlstr(arrayclass, html=0, case=1)

    if verbose:
        print("generateDpage: selectParetoArrays ")
    paretoarrays = oahelper.selectParetoArrays(allarrays, pp)
    at = array2Dtable(paretoarrays, verbose=1)

    if verbose:
        print("generateDpage: write file with Pareto arrays")

    pfile0 = "paretoarrays.oa"
    pfile = os.path.join(outputdir, pfile0)
    oalib.writearrayfile(pfile, paretoarrays)

    istrlnk = markup.oneliner.a("paretoarrays.oa", href=urlprefix + pfile0)

    if lbls is None:
        lbls = ["Optimization of $D$"]
    if fig is not None:
        hh = generateDscatter(dds, lbls=lbls, fig=fig, nofig=nofig)
        oahelper.niceplot(hh.get("ax", None), despine=True, legend=hh["pltlegend"])

        scatterfile = os.path.join(outputdir, "scatterplot.png")
        if verbose:
            print(f"generateDpage: writen scatterplot to {scatterfile}")
        plt.savefig(scatterfile, bbox_inches="tight", pad_inches=0.25, dpi=160)

    page = markup.page()

    if makeheader:
        page.init(
            title=f"Class {xstrplain}",
            css=("../oastyle.css"),
            lang="en",
            htmlattrs=dict({"xmlns": "http://www.w3.org/1999/xhtml", "xml:lang": "en"}),
            header="<!-- Start of page -->",
            bodyattrs=dict({"style": "padding-left: 3px;"}),
            doctype='<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"'
            + ' "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">',
            metainfo=(
                {
                    "text/html": "charset=utf-8",
                    "keywords": "orthogonal arrays designs",
                    "robots": "index, follow",
                    "description": "Even-Odd arrays",
                }
            ),
            footer="<!-- End of page -->",
        )

    page.h1(f"Results for array class {xstr} ")

    ss = r"The Pareto optimaly was calculated according to the statistics D, D<sub>1</sub> and D<sub>s</sub>."
    if npareto == 1:
        page.p("Generated %d arrays, %d is Pareto optimal. %s" % (narrays, npareto, ss))
    else:
        page.p("Generated %d arrays, %d are Pareto optimal. %s" % (narrays, npareto, ss))

    if narrays > 0:
        scores = calcScore(dds, optimfunc)
        _, dd, sols = selectDn(scores, dds, allarrays, nout=1)
        A = sols[0]
        bestdesignfile = os.path.join(outputdir, "best-design.oa")
        oalib.writearrayfile(bestdesignfile, A)

        page.h2("Best design")
        page.p("The best design: {}.".format(e.a("best-design.oa", href=os.path.join(urlprefix, "best-design.oa"))))

        page.p()

        dd = dd[0]
        page.span(f"D-efficiency: {A.Defficiency():.4f}, ")  # page.br()
        page.span(f"D<sub>s</sub>-efficiency: {dd[1]:.4f}, ")  # page.br()
        page.span(f"D<sub>1</sub>-efficiency: {dd[2]:.4f}")
        page.br()
        page.span(f"A-efficiency: {A.Aefficiency():.3f}")
        page.br()
        gwlp = A.GWLP()
        # gwlp=','.join(['%.3f' % xq for xq in gwlp])
        gwlp = oahelper.gwlp2str(gwlp, jstr=", ")
        page.span(f"Generalized wordlength pattern: {gwlp}")
        page.br()
        # page.p('D-efficiency: %.3f' % A.Defficiency() )
        pec = oalib.PECsequence(A)
        pec = ",".join([f"{xq:.3f}" for xq in pec])
        page.span(f"PEC-sequence: {pec}")
        page.br()

    page.h2("Table of Pareto optimal arrays ")

    page.span(str(at))
    page.p(f"All Pareto optimal arrays: {istrlnk}")

    page.img(
        src=urlprefix + "scatterplot.png",
        style="margin: 10px; width:95%; min-width: 300px;  max-width:1100px; height: auto; ",
    )

    citationstr = markup.oneliner.a(
        "Complete Enumeration of Pure-Level and Mixed-Level Orthogonal Arrays", href="https://doi.org/10.1002/jcd.20236"
    )

    page.br(clear="both")
    page.p("Citation notice: if you make use of the results on this page, please cite the following paper:")
    page.p(f"{citationstr}, Journal of Combinatorial Designs, Volume 18, Issue 2, pages 123-140, 2010.")

    page.p(f"Generated with oapackage {oalib.version()}, date {oahelper.timeString()}.")

    outfile = os.path.join(outputdir, "Dresults.html")
    fid = open(outfile, "w")
    fid.write(str(page))
    fid.close()
    print(f"written to file {outfile}")

    return outfile


def scoreDn(dds: FloatArray, optimfunc: list) -> FloatArray:
    """Calculate scores from various efficiencies

    Args:
        dds: calculated D-efficiencies
        optimfunc: parameters for optimization function
    Returns:
        Calculated scores
    """
    scores = np.array([oalib.scoreD(dd, optimfunc) for dd in dds])
    return scores


def calcScore(dds: FloatArray, optimfunc: FloatArray) -> FloatArray:
    """Calculate D-efficiency score using multiple efficiencies and a weight factor

    Args:
        dds (array): the rows contains the effieciencies for various designs
        optimfunc (array): aray with the weight factors for the efficiencies
    """
    scores = np.asarray(dds).dot(np.asarray(optimfunc))

    return scores


def optimDeffPython(
    A0,
    arrayclass=None,
    niter: int = 10000,
    nabort: int = 2500,
    verbose: int = 1,
    alpha: Callable | list[float] = [1, 0, 0],
    method: int = 0,
) -> tuple[FloatArray, Any]:
    """Optimize array using specified optimization method

    Args:
        A0 (array_link): design to optimize
        arrayclass (object): contains class of designs to optimize
        alpha (list): specifies the optimization function

    Returns:
        Tuple with efficiencies and optimized design
    """
    # get factor levels
    if arrayclass is None:
        s = A0.getarray().max(axis=0) + 1
    else:
        s = arrayclass.factor_levels()
    sx = tuple(s.astype(np.int64))
    sx = tuple([int(v) for v in sx])

    sg = oalib.symmetry_group(sx)
    gidx = tuple(sg.gidx)
    gsize = tuple(sg.gsize)
    gstart = tuple(sg.gstart)

    nx = 0

    Dinitial = A0.Defficiency()
    if verbose:
        print(f"optimDeff: initial Deff {Dinitial:.4f}")
    N = A0.n_rows
    k = A0.n_columns

    alpha_is_function = str(type(alpha)) == "<type 'function'>" or callable(alpha)
    # initialize score
    if alpha_is_function:
        efficiencies = alpha(A0)  # type: ignore
    else:
        D, Ds, D1 = A0.Defficiencies()
        efficiencies = alpha[0] * D + alpha[1] * Ds + alpha[2] * D1  # ty: ignore
    A = A0.clone()
    lc = 0
    for ii in range(0, niter):
        r = np.random.randint(N)
        c = np.random.randint(k)
        r2 = np.random.randint(N)
        # make sure c2 is in proper column
        c2 = gstart[gidx[c]] + oalib.fastrand() % gsize[gidx[c]]

        o = A._at(r, c)
        o2 = A._at(r2, c2)  # no extra error checking
        # swap
        if method == oalib.DOPTIM_SWAP:
            # swap
            if o == o2:
                continue
            A._setvalue(r, c, o2)
            A._setvalue(r2, c2, o)
        elif method == oalib.DOPTIM_NONE:
            pass
        else:
            # flip
            A._setvalue(r, c, 1 - o)

        if alpha_is_function:
            dn = alpha(A)  # type: ignore
        else:
            D, Ds, D1 = A.Defficiencies()
            nx = nx + 1
            dn = alpha[0] * D + alpha[1] * Ds + alpha[2] * D1  # type: ignore
        if dn >= efficiencies:
            if dn > efficiencies:
                lc = ii
                efficiencies = dn
            if verbose >= 2:
                print("ii %d: %.6f -> %.6f" % (ii, efficiencies, dn))

        else:
            # restore to original
            if method == oalib.DOPTIM_SWAP:
                A._setvalue(r, c, o)
                A._setvalue(r2, c2, o2)
            elif method == oalib.DOPTIM_NONE:
                pass
            else:
                A._setvalue(r, c, o)
        if (ii - lc) > nabort:
            if verbose:
                print("optimDeff: early abort ii %d, lc %d" % (ii, lc))
            break

    if verbose:
        Dfinal = A.Defficiency()
        if Dfinal > Dinitial:
            print(f"optimDeff: final Deff improved: {Dinitial:.4f} -> {A.Defficiency():.4f}")
        else:
            print(f"optimDeff: final Deff {A.Defficiency():.4f}")

    return efficiencies, A


# %%
def filterPareto(scores, dds, arrays, verbose=0):
    """From a list of designs select only the pareto optimal designs

    Args:
        scores (array): array of scores
        dds (array): array with D-efficiency values
        arrays (list): list of designs
    Returns:
        pareto_scores (list) : list of scores of pareto optimal designs
        pareto_efficiencies (list): list of D-efficiency values of pareto optimal designs
        pareto_designs (list) : list of selected designs
    """
    pp = oahelper.createPareto(dds, verbose=0)
    paretoidx = np.array(pp.allindices())

    pareto_scores = scores[paretoidx]
    pareto_efficiencies = dds[paretoidx]
    pareto_designs = [arrays[i] for i in paretoidx]

    return pareto_scores, pareto_efficiencies, pareto_designs


def selectDn(scores, dds, sols, nout=1, sortfull=True):
    """Select best arrays according to given scores

    The resulting data is sorted

    Parameters
    ----------
    scores : array
        scores for the designs
    dds : array
        array with efficiencies of the designs
    sols : list
        list of designs
    nout : integer or None
        Number of results to return. None means return all results
    sortfull : boolean
        If True, then sort on both the scores and the dds values

    Returns
    -------
    scores, dds, sols : sorted arrays
    """

    if len(sols) == 0:
        return scores, dds, sols

    if sortfull:
        full_sorting_data = [dds[:, ii] for ii in list(range(dds.shape[1]))[::-1]] + [np.array(scores).flatten()]
        idx = np.lexsort(full_sorting_data)[::-1]
        idx = np.lexsort(np.array(full_sorting_data))[::-1]
    else:
        idx = np.argsort(-scores.flatten())
    scores = scores[idx]
    dds = dds[idx, :]
    sols = [sols[ii] for ii in idx]
    if nout is not None:
        # sort the arrays
        nout = np.minimum(nout, scores.size)
        scores = scores[0:nout]
        dds = dds[range(nout), :]
        sols = sols[0:nout]
    return scores, dds, sols


def Doptimize(
    arrayclass,
    nrestarts=10,
    optimfunc=[1, 0, 0],
    verbose=1,
    maxtime=180,
    selectpareto=True,
    nout=None,
    method=oalib.DOPTIM_UPDATE,
    niter=100000,
    nabort=0,
    dverbose=1,
):
    """Calculate D-efficient designs

    The method uses a coordinate-exchange algorithm find a D-efficient (sometimes called D-optimal) design in the
    class specified by the arrayclass. The optimality is defined in terms of the optimization parameters.
    The optimization is performed multiple times (specified by the nrestarts parameter) to prevent finding a design
    in a local minmum of the target function.


    Args:
      arrayclass (object): Specifies the type of design to optimize
      nrestarts (int): Number of restarts of the algorithm
      optimfunc (list with 3 floats): Gives the optimization weights :math:`\\alpha` of the target
                                  function :math:`\\alpha[0] D+\\alpha[1] D_s+\\alpha[2] D_1`
      verbose (int): Verbosity level. A higher numer gives more output
      maxtime (float): Maximum running time of the algorithm. If this time is exceeded the algorithm is aborted.
      selectpareto (bool): default is True. If True then only the Pareto optimal designs are returned
      nout (int or None): Maximum number of designs to return. If None,  return all designs
      method (coordinate_exchange_method_t): Specifies the method use for updating elements
              in the coordinate-exchange algorithm.
      niter (int): Maximum number of iterations of the coordinate-exchange algorithm
      nabort (int): If no improvements have been found after this number of updates, then abort this run
      dverbose (int): Verbosity level passed to the C++ Doptimize function

    Returns
    -------
            scores : list
                list of scores
            dds: array
                array with calculated efficiencies
            designs: list
                list of generated designs
            nrestarts: int
                number of restarts used


    The optimization target and the Pareto optimality are defined in terms of the D-efficiency, main effect robustness
    (or Ds-optimality) and the D1-efficiency of the design. A full definition of these efficiencies is available
    in the documentation
    at https://oapackage.readthedocs.io/en/latest/properties.html#optimality-criteria-for-d-efficient-designs.
    For more details and motivation of these efficiencies, see the
    paper "Two-Level Designs to Estimate All Main Effects and Two-Factor Interactions",
    https://doi.org/10.1080/00401706.2016.1142903.


    """
    if arrayclass.strength != 0:
        warnings.warn("Doptimize can only handle designs with strength 0", UserWarning)

    if verbose:
        print(f"Doptim: optimization class {arrayclass.idstr()}")
    t0 = time.perf_counter()

    if optimfunc is None:
        optimfunc = [1, 2, 0]

    if isinstance(optimfunc, list):
        rr = oalib.Doptimize(
            arrayclass,
            nrestarts,
            alpha=optimfunc,
            verbose=dverbose,
            method=method,
            niter=niter,
            maxtime=maxtime,
            nabort=nabort,
        )
        dds, sols = rr.dds, rr.designs
        dds = np.array([x for x in dds])
        # needed because of SWIG wrapping of struct type
        sols = [design.clone() for design in sols]
        nrestarts = rr.nrestarts
        scores = np.array([oalib.scoreD(A.Defficiencies(), optimfunc) for A in sols])

        if verbose >= 3:
            print(f"Doptimize: max score {np.max(scores):.3f}, max D: {np.max([A.Defficiency() for A in sols]):.6f}")
    else:
        # assume optimfunc is a function
        scores = np.zeros((0, 1))
        dds = np.zeros((0, 3))
        sols = []

        nrestarts_requested = nrestarts
        nrestarts = 0
        for ii in range(nrestarts_requested):
            if verbose:
                oahelper.tprint(
                    "Doptim: iteration %d/%d (time %.1f/%.1f)" % (ii, nrestarts, time.perf_counter() - t0, maxtime),
                    dt=4,
                )
            al = arrayclass.randomarray(1)

            if isinstance(optimfunc, list):
                alpha = optimfunc
                Ax = oalib.optimDeff(al, arrayclass, alpha, verbose >= 2, method, niter, nabort)
                dd = Ax.Defficiencies()
                score = oalib.scoreD(dd, optimfunc)
            else:
                score, Ax = optimDeffPython(al, verbose=0, niter=niter, alpha=optimfunc, method=method, nabort=nabort)
                dd = Ax.Defficiencies()

            if time.perf_counter() - t0 > maxtime:
                if verbose:
                    print("Doptim: max time exceeded, aborting")
                break

            scores = np.vstack((scores, [score]))
            dds = np.vstack((dds, dd))
            sols.append(Ax)
            nrestarts = nrestarts + 1

            if verbose >= 2:
                print(f"  generated array: {dd[0]:f} {dd[1]:f} {dd[2]:f}")

            if selectpareto and ii % 502 == 0:
                scores, dds, sols = filterPareto(scores, dds, sols)

    if selectpareto:
        if verbose >= 2:
            print("Doptim: before Pareto filter (%d arrays)" % len(sols))
        scores, dds, sols = filterPareto(scores, dds, sols)

    if verbose:
        dt = time.perf_counter() - t0
        print("Doptim: done (%d arrays, %.1f [s])" % (len(sols), dt))

    # sort & select
    scores, dds, sols = selectDn(scores, dds, sols, nout=nout)

    return scores, dds, sols, nrestarts



"""
Pieter Eendebak <pieter.eendebak@gmail.com>

"""

# %%
import io
import sys
import unittest
import unittest.mock as mock

import numpy as np

import oapackage
import oapackage.conference


def only_python3(function):
    """Decorator to only execute a test function in Python 3"""
    python3 = sys.version_info >= (3, 4)
    if python3:

        def only_python3_function(*args, **kwargs):
            return function(*args, **kwargs)
    else:

        def only_python3_function(*args, **kwargs):
            return None

    return only_python3_function


# %%


def _statistics_equal(statistics, expected):
    np.testing.assert_array_almost_equal(np.array(statistics), np.array(expected))


class TestFvalues(unittest.TestCase):
    def test_FvaluesConference(self):
        array = oapackage.exampleArray(47, 0)
        js = oapackage.jstructconference_t(array)
        self.assertEqual(js.Jvalues(), (16, 12, 8, 4, 0))
        self.assertEqual(js.calculateF(), (0, 2, 4, 51, 13))

    def test_FvaluesConference_odd_runs(self):
        array = oapackage.exampleArray(47, 0)
        definitive_screening_design = oapackage.conference2DSD(array)
        with self.assertRaises(RuntimeError):
            _ = oapackage.jstructconference_t(definitive_screening_design)


class TestJcharacteristics(unittest.TestCase):
    def test_Jcharacteristics(self):
        array = oapackage.exampleArray(46, 0)

        JJ = oapackage.Jcharacteristics_conference(array, number_of_columns=3)
        self.assertEqual(len(JJ), oapackage.choose(array.n_columns, 3))

        JJ = oapackage.Jcharacteristics_conference(array, 4)
        self.assertEqual(len(JJ), oapackage.choose(array.n_columns, 4))
        hist = np.histogram(np.abs(JJ), bins=np.arange(0, 21, 2))[0]
        self.assertEqual(list(hist), [9, 0, 52, 0, 8, 0, 1, 0, 0, 0])


class TestConferenceStatistics(unittest.TestCase):
    def test_momentMatrix(self):
        M1 = np.array([[1.0, 0.0, 1.0 / 3], [0.0, 1.0 / 3, 0.0], [1.0 / 3, 0.0, 0.2]])
        M = oapackage.conference.momentMatrix(1)
        np.testing.assert_array_equal(M, M1)
        M = oapackage.conference.momentMatrix(4)
        self.assertEqual(M.shape, (15, 15))

    @only_python3
    def test_conference_model_statistics(self):
        arrays = [oapackage.exampleArray(idx, 0).selectFirstColumns(4) for idx in [46, 47, 48]]
        dsds = [oapackage.conference2DSD(array) for array in arrays]
        with mock.patch("sys.stdout", new_callable=io.StringIO) as mock_stdout:
            statistics = [oapackage.conference.modelStatistics(al, verbose=2) for al in dsds]
            stdout = mock_stdout.getvalue()
            self.assertIn("modelStatistics: condition number: ", stdout)

        expected = [(0, 0, 0), (0, 0, 0), (1, 17.081221494279234, 1.7610790263409104)]
        _statistics_equal(statistics, expected)

        arrays = [oapackage.exampleArray(idx, 0).selectLastColumns(4) for idx in [46, 47, 48]]
        dsds = [oapackage.conference2DSD(array) for array in arrays]
        statistics = [oapackage.conference.modelStatistics(al, verbose=0) for al in dsds]
        expected = [
            (1, 17.312358463836613, 1.8538666749267378),
            (1, 17.312358463836613, 1.8538666749267378),
            (1, 17.08122149427922, 1.7610790263409093),
        ]
        _statistics_equal(statistics, expected)

    def test_conference_projection_statistics(self):
        arrays = [oapackage.exampleArray(idx, 0) for idx in [45, 46, 47, 48]]
        statistics = [oapackage.conference.conferenceProjectionStatistics(array, verbose=0) for array in arrays]
        expected = [
            (0.9857142857142858, 16.831420420862223, 1.7523727421346018),
            (0.9857142857142858, 16.782711360159983, 1.7515927378529395),
            (0.9714285714285714, 16.594006884359878, 1.7298655460036123),
            (1.0, 16.80031370680457, 1.7750342960753236),
        ]
        _statistics_equal(statistics, expected)

    def test_leftDivide(self):
        A = np.array([[1, 2], [3, 4]])
        B = np.array([[51, -2], [3, 4]])
        C = oapackage.conference._leftDivide(A, B)
        np.testing.assert_array_almost_equal(C, np.array([[-99.0, 8.0], [75.0, -5.0]]))


if __name__ == "__main__":
    unittest.main()



"""Script to test performance of components of the OA package

Log with:

    > python ptest.py > test0-ddmmyyy.txt
"""

# %% Load necessary packages
import getopt
import os
import platform
import subprocess
import sys
import time

import numpy as np

import oapackage
import oapackage.oahelper as oahelper
from oapackage import oalib

tmp = oalib.log_print(-oalib.SYSTEM, "")

# %% Define test functions


def pytest7(verbose=1):
    """Reduced array with random transformtion to LMC form"""
    al = oalib.exampleArray(3)
    adata = oalib.arraylink2arraydata(al)
    reduction = oalib.LMCreduction_t(adata)

    for ii in range(10):
        if verbose >= 2:
            print("pytest7: randomized reduction of array %d" % ii)
        reduction.transformation.randomize()
        al2 = reduction.transformation.apply(al)

        alr = al2.reduceLMC()
        c = al == alr
        if not c:
            print("pytest7: error: reduction of randomized array failed!")


# %%


def pytest6(verbose=1):
    """Check generation of OA(32, t, 2^a)"""
    N = 32
    k = 8
    t = 3
    l = [2] * 8
    rr = []
    gtrr = [3, 5, 10, 17, 33]
    s = oalib.intVector(l)
    adata = oalib.arraydata_t(s, N, t, k)

    if verbose:
        print("pytest6: run different algorithms on the same case")
    algs = [oalib.MODE_ORIGINAL, oalib.MODE_J4]
    for _, alg in enumerate(algs):
        algname = oalib.algnames(alg)
        if verbose >= 2:
            print(f"pytest6: running {adata.fullidstr()}, alg {algname}")
        rr = []
        oahelper.runExtend(N, k, t, l, verbose=verbose, nums=rr, algorithm=alg)
        if not rr == gtrr:
            print(f"pytest6: case {adata.fullidstr()}")
            print(f"   algorithm {algname}: error: incorrect number of arrays! {rr} -> {gtrr}")


# %%


def pytest3(verbose=2):
    """Test rank and Defficiency properties"""
    k = 6
    strength = 3
    N = 40
    lll = oahelper.runExtend(N, k, strength, 2, verbose=0)

    rnk0 = np.array([19, 19, 21, 21, 19, 22, 22, 22, 22])  # ground-truth
    D0 = np.array(
        [
            0.0,
            0.0,
            0.0,
            0.1703475035645109,
            0.0,
            0.8107439985672059,
            0.8969911485707478,
            0.8953919248315042,
            0.9007069559712273,
        ]
    )  # gt

    rnk = [oalib.array2xf(al).rank() for al in lll]
    D = [al.Defficiency() for al in lll]

    if (rnk - rnk0).any():
        if verbose:
            print("  rank test FAILED!")
            return False
    if not (np.abs(D0 - D) < 1e-6).all():
        if verbose:
            print("  Defficiency test FAILED!")
        return False
    return True


def pytest4(verbose=1):
    """Check generation of OA(16,2, 4^2 2^a)"""
    N = 16
    k = 6
    t = 2
    l = [4, 4, 2]
    rr = []
    oahelper.runExtend(N, k, t, l, verbose=1, nums=rr)
    if not rr == [2, 7, 17, 27]:
        print(f"ERROR: incorrect number of arrays! {rr}")


def pytest2level(verbose=1):
    """Check generation of OA(16,2, 4^2 2^a)"""
    N = 32
    k = 10
    t = 3
    l = [2]
    rr = []
    t0 = time.time()
    oahelper.runExtend(N, k, t, l, verbose=1, nums=rr, algorithm=oalib.MODE_ORIGINAL)
    dt = time.time() - t0
    t0 = time.time()
    oahelper.runExtend(N, k, t, l, verbose=1, nums=rr, algorithm=oalib.MODE_LMC_2LEVEL)
    dt2level = time.time() - t0

    ll = l * k
    s = oalib.intVector(ll)

    adata = oalib.arraydata_t(s, N, t, k)

    if verbose:
        print(f"case {adata.idstr()}: 2-level method {dt:.2f} [s] -> {dt2level:.2f} [s]")
    # cases.append('runExtend(N=%d, k=%d, t=%d, l=%s)' %(N,k,t, str(l)) )


def pytest_analysis(oadir, verbose=1):
    """Performance of analysis algorithms"""
    cmd = "export OMP_NUM_THREADS=1; oaanalyse --gwp -A {}".format(os.path.join(oadir, "testdata", "test64.oa"))
    if verbose:
        print("pytest_analysis: running oaanalyse")
    t0 = time.time()
    return_value = os.system(cmd)
    assert return_value == 0
    ta = time.time() - t0
    return ta


def pytest2(oadir, verbose=1):
    afile = os.path.join(oadir, "testdata", "unstable-rank.oa")
    sols = oalib.readarrayfile(afile)
    afile = os.path.join(oadir, "testdata", "unstable-rank-extended.oa")
    sols = oalib.readarrayfile(afile)

    for ii, al in enumerate(sols):
        vv = oalib.doubleVector(3)
        rnk = oalib.array2rank_Deff_Beff(al, vv)

        al2 = oalib.array2xf(al)
        rnk2 = al2.rank()
        if (not rnk2 == rnk) or (not rnk < 29):
            print("warning: problem?!: %d %d %d" % (ii, rnk, rnk2))
        if verbose >= 2:
            print("array %d: rank: %d %d" % (ii, rnk, rnk2))


def pytest(verbose=1):
    t0 = time.time()
    tt = []
    cases = []

    N = 48
    k = 6
    t = 3
    l = 2
    rr = []
    oahelper.runExtend(N, k, t, l, verbose=1, nums=rr)
    tt.append(time.time() - t0)
    cases.append("runExtend(N=%d, k=%d, t=%d, l=%d)" % (N, k, t, l))
    if not rr == [4, 10, 45]:
        print(f"ERROR: incorrect number of arrays! {rr}")

    N = 32
    k = 6
    t = 3
    l = 2
    rr = []
    oahelper.runExtend(N, k, t, l, verbose=1, nums=rr)
    tt.append(time.time() - t0)
    cases.append("runExtend(N=%d, k=%d, t=%d, l=%d)" % (N, k, t, l))
    if not rr == [3, 5, 10]:
        print(f"ERROR: incorrect number of arrays! {rr}")

    N = 16
    k = 16
    t = 2
    l = 2
    rr = []
    oahelper.runExtend(N, k, t, l, verbose=1, nums=rr)
    tt.append(time.time() - t0)
    cases.append("runExtend(N=%d, k=%d, t=%d, l=%d)" % (N, k, t, l))
    if not rr == [3, 5, 11, 27, 55, 80, 87, 78, 58, 36, 18, 10, 5, 0]:
        print(f"ERROR: incorrect number of arrays! {rr}")

    N = 96
    k = 8
    t = 4
    l = 2
    rr = []
    oahelper.runExtend(N, k, t, l, verbose=1, nums=rr)
    tt.append(time.time() - t0)
    cases.append("runExtend(N=%d, k=%d, t=%d, l=%d)" % (N, k, t, l))
    if not rr == [4, 9, 4, 0]:
        print(f"ERROR: incorrect number of arrays! {rr}")

    dt = time.time() - t0
    if verbose:
        print(f"total time {dt:.3f} [s]")

    if verbose:
        print("extra testing cases")
    N = 24
    k = 5
    t = 2
    l = [3, 2]
    rr = []
    oahelper.runExtend(N, k, t, l, verbose=1, nums=rr)
    tt.append(time.time() - t0)
    cases.append("runExtend(N=%d, k=%d, t=%d, l=%s)" % (N, k, t, str(l)))
    if not rr == [4, 29, 573]:
        print(f"ERROR: incorrect number of arrays! {rr}")

    N = 18
    k = 9
    t = 2
    # old test: l=[2,3]
    l = [3] * 8 + [2]
    rr = []
    oahelper.runExtend(N, k, t, l, verbose=1, nums=rr)
    tt.append(time.time() - t0)
    cases.append("runExtend(N=%d, k=%d, t=%d, l=%s)" % (N, k, t, str(l)))
    if not rr == [4, 12, 10, 8, 3, 0, 0]:
        print(f"ERROR: incorrect number of arrays! {rr}")

    dt2 = time.time() - t0
    return (dt, tt, cases, dt2)


class Usage(Exception):
    def __init__(self, msg):
        self.msg = msg


# %%


def testExtendBinary(verbose=1):
    repos = dict()
    repos["oatest1.txt"] = "result-16.2-2-2-2-2-2-2-2.oa"
    repos["oatest2.txt"] = "result-18.3-3-3-3-3.oa"
    repos["oatest3.txt"] = "result-8.4-2-1-4-2-2.oa"
    repos["oatest4.txt"] = "result-28.7-2-2-2.oa"
    repos["oatest5.txt"] = "result-16.2-2-2-2-2-2-2-2-2.oa"
    repos["oatest6.txt"] = "result-25.5-5-5-5.oa"
    repos["oatest7.txt"] = "result-64.2-2-2-2-2-2.oa"
    repos["oatest8.txt"] = "result-56.2-2-2-2-2.oa"


# testExtendBinary()


# %%
def main(argv=None):
    """Main testing function"""
    if platform.system() == "Windows":
        oadir = os.path.join(os.path.split(oapackage.__file__)[0], "..")
    else:
        oadir = os.path.join(os.path.expanduser("~"), "misc/oa/oacode/")

    print("OA performance testing: version 2.0")
    ss = oapackage.version()
    print(f"OAlib: version {ss}")
    ss = oalib.compile_information()
    print(ss)
    print("System:")
    print("  Python: " + " ".join(sys.version.split("\n")))
    print("  Machine: " + " ".join([f"{x}" for x in platform.uname()]))
    print(f"  Processor: {platform.processor()}")

    if argv is None:
        argv = sys.argv
    try:
        try:
            opts, args = getopt.getopt(argv[1:], "h", ["help"])
        except getopt.error as msg:
            raise Usage(msg)
        # more code, unchanged
    except Usage as err:
        print >> sys.stderr, err.msg
        print >> sys.stderr, "for help use --help"
        return 2

    print("\nRunning tests:\n")
    (dt, tt, cases, dt2) = pytest()

    pytest2(oadir)
    pytest3(verbose=0)
    pytest4(verbose=1)
    ta1 = pytest_analysis(oadir, verbose=1)
    pytest6(verbose=0)
    pytest7(verbose=0)

    pytest2level(verbose=1)

    print("-----")

    dtx = ta1

    print("\nResults:\n")
    for ii, t in enumerate(tt):
        print(f"{cases[ii]}: time {t:.2f} [s]")
    version = oapackage.__version__
    print(f"Total time: {dt:.3f} [s], {dt2:.3f} [s], {dtx:.3f} [s] ({platform.node()}, {version})")
    print("   should be of order 3.408 [s], 3.494 [s], 7.313 [s] (woelmuis, 2.4.8)")
    print("   should be of order 4.4 [s], 4.6 [s], 5.9 [s] (woelmuis)")
    print("   should be of order 3.2 [s], 3.3 [s], 4.7 [s] (marmot) [v 2.0.0]")
    print("   should be of order 2.96 [s], 3.0 [s], 4.6 [s] (marmot) [v 2.0.24]")


if __name__ == "__main__":
    main()


def timeconfig(
    configfile="oaconfig.txt", timebin="/usr/bin/time", pdir="/home/eendebakpt/misc/oa/oacode/performancetest"
):
    os.chdir(pdir)
    res = subprocess.check_output(
        [(f"cd {pdir};ls;") + timebin, '--format="%%E %%S %%U"', "./oaextendsingle", f"-c {configfile} -l 1"]
    )
    return res



import oapackage._scanf


def sscanf(inputString, formatString):
    """Simple scanf function

    Args:
        inputString (str): string to be parsed
        formatString (str): specification
    Returns:
        list: list of parsed arguments
    """
    return oapackage._scanf.scanf(formatString, inputString)



"""Orthogonal Array package test functions"""

import importlib
import os
import tempfile
import unittest
import unittest.mock as mock
from unittest.mock import patch

import numpy as np
import oalib

import oapackage
import oapackage.graphtools
import oapackage.scanf
from oapackage.oahelper import write_text_arrayfile


def autodoctest():
    """Test the module using autodoc
    Example:
      >>> import oapackage
      >>> arrayclass=oapackage.arraydata_t(2, 40, 0, 7)
      >>> print(arrayclass)
      arrayclass: N 40, k 7, strength 0, s {2,2,2,2,2,2,2}, order 0

    """
    return


class TestMisc(unittest.TestCase):
    def test_reduceGraphNauty(self):
        G = np.zeros((5, 5), dtype=int)
        G[1, 0] = G[0, 1] = 1
        v = oapackage.reduceGraphNauty(G)
        self.assertTrue(len(v) == G.shape[0])

    def test_exampleArray(self):
        # test a selection of the example arrays
        al = oapackage.exampleArray(5, 0)
        self.assertTrue(al.md5() == "3885c883d3bee0c7546511255bb5c3ae")
        al = oapackage.exampleArray(20, 0)
        self.assertTrue(np.array(al).shape == (24, 3))
        al = oapackage.exampleArray(40, 0)
        self.assertTrue(np.array(al).shape == (14, 5))

    def test_scanf(self):
        r = oapackage.scanf.sscanf("1", "%d")
        self.assertTrue(r[0] == 1)

    def test_oa2graph(self):
        al = oapackage.exampleArray(2, 0)
        adata = oapackage.arraylink2arraydata(al)
        g = oapackage.graphtools.oa2graph(al, adata)
        self.assertTrue(g[0].shape == (34, 34))


def test_extendSingleArray():
    A = oapackage.exampleArray(4, 1)
    adata = oapackage.arraylink2arraydata(A, extracols=2)
    B = A.selectFirstColumns(5)
    ee = oapackage.oahelper.extendSingleArray(B, adata, t=2, verbose=1)
    assert ee[0].n_columns == B.n_columns + 1
    assert ee[1] == A.selectFirstColumns(6)


def test_numpy_interface(verbose=0):
    A = np.eye(3, 4).astype(int)
    A[0, :] = [10, 20, 30, 50]
    A[2, :] = [-3, -4, -5, -6]

    if verbose:
        print("makearraylink")
    al = oapackage.makearraylink(A)

    np.testing.assert_array_equal(np.array(al), A)

    if verbose:
        al.showarray()

    if verbose:
        print("direct")
    al = oapackage.array_link(A)
    if verbose:
        al.showarray()
    Ax = np.array(al)
    if verbose:
        print(A)
        print(Ax)

    with np.testing.assert_raises(TypeError):
        # not possible right now...
        if verbose:
            print("direct float")
        A = np.eye(3).astype(float)
        al = oapackage.array_link(A)
        if verbose:
            al.showarray()


def test_nauty(verbose=0):
    if verbose:
        print("test_nauty: test reduction to normal form")
    al = oapackage.exampleArray(0, verbose)
    tr = oapackage.reduceOAnauty(al)
    al_reduced = tr.apply(al)

    alrandom = al.randomperm()
    tr = oapackage.reduceOAnauty(alrandom)
    al_reduced2 = tr.apply(alrandom)

    assert al_reduced == al_reduced2


def miscunittest(verbose=1):
    """Perform some unit testing, return True if succesfull"""
    if verbose:
        print(f"oapackage: unittest: oalib version {oalib.version()}")
    al = oalib.array_link()
    ii = 0
    al = oalib.exampleArray(ii, 0)

    arrayclass = oalib.arraydata_t(2, 16, 2, 6)  # fine
    alist = [arrayclass.create_root()]
    r = oalib.extend_arraylist(alist, arrayclass)
    if len(r) != 3:
        raise Exception(f"extension generation for {arrayclass} failed")

    if not isinstance(al.getarray(), np.ndarray):
        print("oapackage: unittest: error: array interface not working properly")
    else:
        if not al[2, 0] == al.getarray()[2, 0]:
            print("oapackage: unittest: error: array interface not working properly")

    arrayclass = oalib.arraylink2arraydata(al)

    if verbose >= 2:
        print("unittest: calculate efficiencies")
    Deff = al.Defficiency()
    aa = oalib.Aefficiencies(al)
    assert aa[0] == 1.0
    assert aa[1] == 1.0
    assert aa[2] == 1.0

    if verbose >= 2:
        print(f"## oapackage test: example array {ii}: Deff {Deff:%.3f}")

    # DOP reduction
    if verbose >= 2:
        print("unittest: test delete-one-factor GWLP reduction")
    al = oalib.exampleArray(5, verbose)
    al2 = al.randomperm()

    alr = al.reduceDOP()
    al2r = al2.reduceDOP()
    if not alr == al2r:
        print(f"error: DOP reduced arrays unequal!: {alr == al2r}")
        print("alr")
        alr.showarraycompact()
        print("al2r")
        al2r.showarraycompact()
        return False

    transformation = oalib.reductionDOP(al)
    check = transformation.apply(al) == al.reduceDOP()
    if not check:
        print("error: DOP reduction transformation is invalid")

    # test graphtools
    if verbose >= 2:
        print("unittest: test graphtools")
    from oapackage.graphtools import oa2graph

    arrayclass = oalib.arraylink2arraydata(al)
    _ = oa2graph(al, arrayclass)

    test_numpy_interface()
    test_nauty()

    return True


# %%


class TestOAfiles(unittest.TestCase):
    """Test functionality related to orthogonal array files"""

    def test_misc_file_operations(self):
        _, array_filename = tempfile.mkstemp(suffix=".oa")
        lst = [oapackage.exampleArray(4, 0)]
        oapackage.writearrayfile(array_filename, lst)
        assert oapackage.oahelper.oaIsBinary(array_filename) is False
        oapackage.writearrayfile(array_filename, oapackage.exampleArray(4, 0), oapackage.ABINARY)
        assert oapackage.oahelper.oaIsBinary(array_filename)

        oapackage.oahelper.oainfo(array_filename)

        _ = oapackage.oahelper.compressOAfile(array_filename)

    def test_selectArrays(self):
        _, array_filename = tempfile.mkstemp(suffix=".oa", dir=tempfile.tempdir)
        _, array_filename_out = tempfile.mkstemp(suffix=".oa", dir=tempfile.tempdir)
        oapackage.writearrayfile(array_filename, [oapackage.exampleArray(4, 0), oapackage.exampleArray(4, 0)])
        oapackage.oahelper.selectArraysInFile(
            array_filename, array_filename_out, [1], afmode=oalib.ABINARY, verbose=1, cache=0
        )
        oapackage.oahelper.selectArraysInFile(
            array_filename, array_filename_out, [-1], afmode=oalib.ABINARY, verbose=1, cache=0
        )

    def test_write_arrayfile_with_comment(self):
        designs = [oapackage.exampleArray(3)]
        filename = tempfile.mktemp(suffix=".oa")
        write_text_arrayfile(filename, designs, comment="Test comment")
        arrays = oapackage.readarrayfile(filename, 0)
        self.assertEqual(len(designs), len(arrays))
        self.assertEqual(designs[0], arrays[0])

    def test_nArrayFile(self):
        _, array_filename = tempfile.mkstemp(suffix=".oa", dir=tempfile.tempdir)
        oapackage.writearrayfile(array_filename, [oapackage.exampleArray(4, 0)])
        self.assertEqual(oapackage.oahelper.nArrayFile(array_filename), 1)
        self.assertEqual(oapackage.oahelper.nArrayFile("notavalidfile.oa"), -1)

    def test_findfiles(self):
        _, array_filename = tempfile.mkstemp(suffix=".oa", dir=tempfile.tempdir)
        lst = [oapackage.exampleArray(4, 0)]
        oapackage.writearrayfile(array_filename, lst)
        lst = oapackage.oahelper.findfiles(tempfile.tempdir, ".*oa")
        self.assertIn(os.path.split(array_filename)[-1], lst)

    def test_findfilesR(self):
        _ = oapackage.oahelper.findfilesR(tempfile.tempdir, ".*oa")

    def test_checkArrayFile(self):
        a = tempfile.mktemp(suffix=".oa")
        self.assertFalse(oapackage.oahelper.checkArrayFile(a))
        self.assertTrue(oapackage.oahelper.checkArrayFile(a, -1))

    def test_finddirectories(self):
        _ = oapackage.oahelper.finddirectories(os.getcwd())


class TestParetoFunctionality:
    def test_selectParetoArrays(self):
        arrays = [oapackage.array_link(np.array([[ii]])) for ii in range(5)]
        pareto_object = oapackage.ParetoLongLong()

        for ii in range(len(arrays)):
            value = [ii, ii % 2]
            pareto_object.addvalue(value, ii)
        pareto_object.show(2)

        selected = oapackage.oahelper.selectParetoArrays(arrays, pareto_object)
        self.assertEqual(selected, arrays[4, 5])


class TestOAhinterface(unittest.TestCase):
    """Test functionality in C++ to Python interface"""

    def test_update_array_link(self):
        al = oapackage.exampleArray(1)
        oapackage.update_array_link(al, np.array([[1, 2, 3], [4, 5, 6]]))
        print(al)
        self.assertEqual(al.shape, (2, 3))
        self.assertEqual(list(al), [1, 4, 2, 5, 3, 6])


class TestOAhelper(unittest.TestCase):
    """Test functionality contained in oahelper module"""

    def setUp(self):
        self.test_array = oapackage.exampleArray(1, 0)

    def test_helmert_contrasts(self):
        hc = oapackage.oahelper.helmert_contrasts(2, verbose=0)
        np.testing.assert_array_almost_equal(hc, np.array([[-1.0], [1.0]]))

        hc = oapackage.oahelper.helmert_contrasts(3, verbose=0)
        np.testing.assert_array_almost_equal(
            hc, np.array([[-1.22474487, -0.70710678], [1.22474487, -0.70710678], [0.0, 1.41421356]])
        )

        hc = oapackage.oahelper.helmert_contrasts(10, verbose=0)
        np.testing.assert_array_almost_equal(
            hc[0],
            np.array(
                [
                    -2.23606798,
                    -1.29099445,
                    -0.91287093,
                    -0.70710678,
                    -0.57735027,
                    -0.48795004,
                    -0.42257713,
                    -0.372678,
                    -0.33333333,
                ]
            ),
        )

        for num_levels in [4, 10, 12]:
            hc = oapackage.oahelper.helmert_contrasts(num_levels, verbose=0)
            np.testing.assert_array_almost_equal(hc.T.dot(hc), num_levels * np.eye(num_levels - 1))

    def test_array2latex(self):
        latex_str = oapackage.oahelper.array2latex(np.array(self.test_array))
        self.assertEqual(latex_str[0:15], r"\begin{tabular}")
        latex_str = oapackage.oahelper.array2latex(np.array(self.test_array), mode="psmallmatrix")
        self.assertEqual(latex_str[0:15], r"\begin{psmallma")
        latex_str = oapackage.oahelper.array2latex(np.array(self.test_array), mode="pmatrix")
        self.assertEqual(latex_str[0:15], r"\begin{pmatrix}")

    def test_gwlp2str(self):
        with self.assertWarns(UserWarning):
            self.assertEqual(oapackage.oahelper.gwlp2str([1, 2, 3]), "")
        self.assertEqual(oapackage.oahelper.gwlp2str([1, 0, 0.1]), "1.00,0.00,0.10")

    def test_parseProcessingTime(self):
        filename = tempfile.mktemp(suffix=".txt")
        with open(filename, "w") as fid:
            fid.write("# my logfile\n")
            fid.write("#time start: 2012-01-19 17:21:00\n")
            fid.write("#time end: 2012-01-19 18:21:00\n")
        dtt = oapackage.oahelper.parseProcessingTime(filename, verbose=2)
        self.assertEqual(dtt, 3600.0)

    def test_argsort(self):
        idx = oapackage.oahelper.argsort([1, 2, 3])
        assert idx == [0, 1, 2]
        idx = oapackage.oahelper.argsort([2, 2, 1])
        assert idx == [2, 0, 1]

    def test_plot2Dline(self):
        if importlib.util.find_spec("matplotlib") is not None:
            with mock.patch("matplotlib.pyplot.ylim"):
                with mock.patch("matplotlib.pyplot.plot") as MockPlt:
                    oapackage.oahelper.plot2Dline([1, 0, 0])
                    self.assertTrue(MockPlt.called)

    def test_deprecated(self):
        def func():
            return "hi"

        deprecated_function = oapackage.oahelper.deprecated(func)
        with self.assertWarns(Warning):
            _ = deprecated_function()

    def test_formatC(self):
        c_code = oapackage.oahelper.formatC(self.test_array)
        self.assertEqual(
            c_code,
            (
                "\tarray_link array ( 16,5, 0 );\n\tint array_data_tmp[] = "
                "{0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,0,0,0,0,1,1,1,1,0,0,0,1,0,1,1,1,0,1,1,1,0,0,0,1,0,0,1,0,1,0,"
                "1,1,1,0,1,1,0,0,1,0,0,0,1,1,0,0,1,1,1,1,0,0,1,1,0,0};"
                "\tarray.setarraydata(array_data_tmp, array.n_rows * array.n_columns);\n"
            ),
        )

    def test_runExtend(self):
        N = 24
        k = 5
        t = 2
        r = oapackage.oahelper.runExtend(N, 4, t=t, l=2, verbose=3)
        rx = oapackage.oahelper.runExtend(N, k, t=t, l=2, initsols=r, verbose=3)
        self.assertTrue(len(r) == 10)
        self.assertTrue(len(rx) == 63)

    def test_joinArrayLists(self):
        l1 = [oapackage.exampleArray(2, 0)]
        l2 = [oapackage.exampleArray(2, 0), oapackage.exampleArray(2, 0)]
        joined_list = oapackage.oahelper.joinArrayLists([l1, l2])
        assert len(joined_list) == len(l1) + len(l2)

    def test_checkFiles(self):
        self.assertTrue(oapackage.oahelper.checkFiles([], -1))
        self.assertTrue(oapackage.oahelper.checkFiles(["nosuchfile"], -1))
        self.assertTrue(oapackage.oahelper.checkFiles("nosuchfile", -1))

    def test_checkFilesOA(self):
        r = oapackage.oahelper.checkFilesOA([], cache=1, verbose=0)
        self.assertTrue(r)

    def test_floatformat(self):
        s = oapackage.oahelper.floatformat(3.14, mind=2, maxd=4)
        self.assertEqual(s, "3.14")
        s = oapackage.oahelper.floatformat(3.14, mind=1, maxd=2)
        self.assertEqual(s, "3.1")

    def test_safemin_safemax(self):
        r = oapackage.oahelper.safemin(np.array([1, -2, 3]), default=0)
        self.assertEqual(r, -2)
        r = oapackage.oahelper.safemin(np.array([]), default=-2)
        self.assertEqual(r, -2)
        r = oapackage.oahelper.safemax(np.array([1, -2, 3]), default=0)
        self.assertEqual(r, 3)
        r = oapackage.oahelper.safemax([], default=-23)
        self.assertEqual(r, -23)

    def test_choose(self):
        test_cases = [((3, 2), 3), ((10, 1), 10), ((5, 2), 10), ((1, 0), 1), ((-1, 0), 1)]
        for args, expected in test_cases:
            result = oapackage.oahelper.choose(*args)
            self.assertEqual(result, expected)

    def test_create_pareto_element(self):
        values = [1, 2, 3]
        p = oapackage.oahelper.create_pareto_element(values, pareto=None)
        self.assertEqual(p, values)

        self.assertRaises(Exception, oapackage.oahelper.create_pareto_element, dict())

        pareto = oalib.ParetoMultiDoubleLong()
        pareto_element = oapackage.oahelper.create_pareto_element([[1, 2], [3, 4]], pareto)
        self.assertEqual(pareto_element.size(), 2)

        pareto = oalib.ParetoMultiLongLong()
        pareto_element = oapackage.oahelper.create_pareto_element([[1, 2], [3, 4]], pareto)
        self.assertEqual(pareto_element.size(), 2)

    def test_create_pareto_element_invalid_type(self):
        with self.assertRaises(Exception):
            oapackage.oahelper.create_pareto_element([1.0], 1.0)

    def test_designStandardError(self):
        al = oapackage.exampleArray(14, 0)
        v = oapackage.oahelper.designStandardError(al)
        self.assertAlmostEqual(v[0], 0.3747931073686535)
        al = oapackage.exampleArray(9, 0)
        v = oapackage.oahelper.designStandardError(al)
        np.testing.assert_array_almost_equal(
            v[1], np.array([0.1679305, 0.17229075, 0.17286095, 0.17287786, 0.17303912, 0.17353519, 0.17548291])
        )

    def test_fac(self):
        self.assertEqual(oapackage.oahelper.fac(4), 24)

    def test_testHtml(self):
        import webbrowser

        with patch.object(webbrowser, "open", return_value=None):
            oapackage.oahelper.testHtml("<p>hi</p>")

    def test_bounds(self):
        b = oapackage.oahelper.DefficiencyBound(0.8, 4, 6)
        self.assertAlmostEqual(b, 0.8944271909999)

    def test_array2html(self):
        X = np.array([[1, 2], [3, 4]], dtype="U200")
        h = oapackage.oahelper.array2html(
            X, header=1, tablestyle="border-collapse: collapse;", trclass="", tdstyle="", trstyle="", thstyle=""
        )
        assert "table" in str(h)

    def test_sortrows(self):
        a = np.array([[1, 1], [-2, 2], [-3, 3], [-4, 4], [5, 5]])
        s = oapackage.oahelper.sortrows(a)
        self.assertTrue(np.all(s == [3, 2, 1, 0, 4]))

        x = -np.array([[0, 0, 1], [0, 1, 0], [0, 1, 1]])
        idx = oapackage.oahelper.sortrows(x)
        assert np.all(idx == np.array([2, 1, 0], dtype=np.int64))

        x = np.array([])
        idx = oapackage.oahelper.sortrows(x)
        self.assertTrue(len(idx) == 0)

    def test_sortcols(self):
        a = np.array([[1, 1], [-2, 2], [-3, 3], [-4, 4], [5, 5]]).T
        s = oapackage.oahelper.sortcols(a)
        self.assertTrue(np.all(s == [3, 2, 1, 0, 4]))


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np

import oapackage


class TestMacWilliams(unittest.TestCase):
    def test_macwilliams_transform(self):
        array = oapackage.exampleArray(1, 0)
        array = array.selectFirstColumns(3)
        arrayclass = oapackage.arraylink2arraydata(array)

        Dm = oapackage.distance_distribution_mixed(array, 0)
        D = np.array(Dm)

        N = array.n_rows
        s = 2
        factor_levels_for_groups = arrayclass.factor_levels_column_groups()

        Bp = oapackage.macwilliams_transform(D, N, s)
        Bpm = oapackage.macwilliams_transform_mixed(Dm, N, factor_levels_for_groups, verbose=0)

        np.testing.assert_array_equal(Bp, [1, 0, 0, 0.25])
        np.testing.assert_array_equal(Bp, Bpm)

        oapackage.GWLP(array)
        oapackage.GWLPmixed(array, verbose=2)


if __name__ == "__main__":
    """ Test code """
    unittest.main()



#!/usr/bin/python
"""

Example script for Python interface to Orthogonal Array code.

mathescape, numbersep=5pt, label=Python session, gobble=0, frame=lines, framesep=1.5mm, bgcolor=bg

> pygmentize -P "bgcolor=.95,.95,.95" -P "style=colorful" -f html example_python_testing.py
> pygmentize -S colorful -f html > style.css; cp style.css ~/misc/homepage/oapage/


@author: Pieter Eendebak
"""

# import oalib
import oapackage

print(f"oalib version: {oapackage.version()}")

al = oapackage.exampleArray(0)
al.showarray()
print("D-efficiency %f, rank %d" % (al.Defficiency(), al.rank()))
print(f"Generalized wordlength pattern: {str(al.GWLP())}")


al = oapackage.exampleArray(1)
print(f"Generalized wordlength pattern: {str(al.GWLP())}")



"""Module with functions to work with conference designs"""

import itertools

import numpy as np

import oapackage

# %%


def momentMatrix(k: int) -> np.ndarray:
    """Return the moment matrix of a conference design

    Args:
        k (int): number of columns of the conference design
    Returns:
        array: moment matrix
    """
    pk = int(1 + 0.5 * k * (k + 1) + k)
    M = np.zeros((pk, pk))
    M[0, 0] = 1
    M[0, int(0.5 * k * (k + 1) + 1) :] = 1.0 / 3
    M[int(0.5 * k * (k + 1) + 1) :, 0] = 1.0 / 3
    M[1 : (k + 1), 1 : (k + 1)] = np.eye(k) * 1.0 / 3
    M[(k + 1) : int(0.5 * k * (k + 1) + 1), (k + 1) : int(0.5 * k * (k + 1) + 1)] = np.eye(int(0.5 * k * (k - 1))) / 9.0
    M[int(0.5 * k * (k + 1) + 1) :, int(0.5 * k * (k + 1) + 1) :] = np.eye(k) / 5 + (np.ones((k, k)) - np.eye(k)) / 9.0
    return M


def _leftDivide(A, B):
    r"""Perform left division of a matrix

    Args:
        A (aray)
        B (array)
    Returns:
        array: the result of A\B
    """
    solution = np.linalg.solve(A.T.dot(A), A.T.dot(B))
    return solution


def modelStatistics(dsd, verbose=0, moment_matrix=None, use_condition_number=True):
    """Calculate statistics of a definitive screening design from the model matrix

    Args:
        dsd (array): definitive screening design
    Returns:
      list: calculated statistics. Calculated are whether the model is estible, the Defficiency and the inverse APV.
    """
    ncolumns = dsd.shape[1]
    modelmatrix = np.array(oapackage.conference_design2modelmatrix(dsd, "q", verbose=0))
    M = (modelmatrix.T).dot(modelmatrix)

    if use_condition_number:
        fullrank = np.linalg.cond(M) < 1000
    else:
        mr = np.linalg.matrix_rank(M)
        fullrank = mr == modelmatrix.shape[1]
        if verbose >= 2:
            print(
                "modelStatistics: fullrank %d, condition number %.4f"
                % (
                    fullrank,
                    np.linalg.cond(M),
                )
            )

    if verbose >= 2:
        print(f"modelStatistics: condition number: {np.linalg.cond(M)}")
    if fullrank:
        if moment_matrix is None:
            moment_matrix = momentMatrix(ncolumns)
        Eest = 1
        pk = int(1 + ncolumns + ncolumns * (ncolumns + 1) / 2)
        kappa = np.linalg.det(M)
        lnkappa = np.log(kappa) / (pk)
        Defficiency = np.exp(lnkappa)

        apv = np.trace(_leftDivide(M, moment_matrix))
        invAPV = 1 / apv
    else:
        Eest = 0
        Defficiency = 0
        invAPV = 0

    return Eest, Defficiency, invAPV


def conferenceProjectionStatistics(array, ncolumns=4, verbose=0):
    """Calculate the projection statistics of a conference design

    The PECk, PICk and PPCk are calculated with k the number of columns specified.
    The projection statistics are calculated by determined the :meth:`modelStatistics` of all k-column subdesigns
    and then taking the mean of the statistics for the subdesigns.
    For more details of the calculation, see https://oapackage.readthedocs.io/.

    Args:
        array (array): conference design
        ncolumns (int): number of columns on which to project

    Returns:
        pec, pic, ppc (float): calculated statistics
    """
    nc = array.shape[1]

    number_combinations = oapackage.choose(nc, ncolumns)
    Eestx = np.zeros(number_combinations)
    Deff = np.zeros(number_combinations)
    invAPV_values = np.zeros(number_combinations)
    dsd = oapackage.conference2DSD(oapackage.array_link(array))
    moment_matrix = momentMatrix(ncolumns)
    for idx, c in enumerate(list(itertools.combinations(range(nc), ncolumns))):
        proj_dsd = dsd.selectColumns(c)
        Eest, D, invAPV = modelStatistics(proj_dsd, verbose=0, moment_matrix=moment_matrix)

        Deff[idx] = D
        Eestx[idx] = Eest
        invAPV_values[idx] = invAPV
    pec, pic, ppc = np.mean(Eestx), np.mean(Deff), np.mean(invAPV_values)
    if verbose:
        print(
            "conferenceProjectionStatistics: projection to %d columns: PEC %.3f PIC %.3f PPC %.3f  "
            % (ncolumns, pec, pic, ppc)
        )
    return pec, pic, ppc
